# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the class
:class:`iris.fileformats.um._fast_load_structured_fields.BasicFieldCollation`.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

from iris._lazy_data import as_lazy_data
from cftime import datetime
import numpy as np

from iris.fileformats.um._fast_load_structured_fields import (
    BasicFieldCollation,
)
import iris.fileformats.pp


class Test___init__(tests.IrisTest):
    def test_no_fields(self):
        with self.assertRaises(AssertionError):
            BasicFieldCollation([])


class Test_fields(tests.IrisTest):
    def test_preserve_members(self):
        fields = ("foo", "bar", "wibble")
        collation = BasicFieldCollation(fields)
        self.assertEqual(collation.fields, fields)


def _make_field(
    lbyr=None, lbyrd=None, lbft=None, blev=None, bhlev=None, data=None
):
    header = [0] * 64
    if lbyr is not None:
        header[0] = lbyr
        header[1] = 1
        header[2] = 1
    if lbyrd is not None:
        header[6] = lbyrd
        header[7] = 1
        header[8] = 1
    if lbft is not None:
        header[13] = lbft
    if blev is not None:
        header[51] = blev
    if bhlev is not None:
        header[53] = bhlev
    field = iris.fileformats.pp.PPField3(header)
    if data is not None:
        _data = _make_data(data)
        field.data = _data
    return field


def _make_data(fill_value):
    shape = (10, 10)
    return as_lazy_data(np.ones(shape) * fill_value)


class Test_data(tests.IrisTest):
    # Test order of the data attribute when fastest-varying element is changed.
    def test_t1_varies_faster(self):
        collation = BasicFieldCollation(
            [
                _make_field(lbyr=2013, lbyrd=2000, data=0),
                _make_field(lbyr=2014, lbyrd=2000, data=1),
                _make_field(lbyr=2015, lbyrd=2000, data=2),
                _make_field(lbyr=2013, lbyrd=2001, data=3),
                _make_field(lbyr=2014, lbyrd=2001, data=4),
                _make_field(lbyr=2015, lbyrd=2001, data=5),
            ]
        )
        result = collation.data[:, :, 0, 0]
        expected = [[0, 1, 2], [3, 4, 5]]
        self.assertArrayEqual(result, expected)

    def test_t2_varies_faster(self):
        collation = BasicFieldCollation(
            [
                _make_field(lbyr=2013, lbyrd=2000, data=0),
                _make_field(lbyr=2013, lbyrd=2001, data=1),
                _make_field(lbyr=2013, lbyrd=2002, data=2),
                _make_field(lbyr=2014, lbyrd=2000, data=3),
                _make_field(lbyr=2014, lbyrd=2001, data=4),
                _make_field(lbyr=2014, lbyrd=2002, data=5),
            ]
        )
        result = collation.data[:, :, 0, 0]
        expected = [[0, 1, 2], [3, 4, 5]]
        self.assertArrayEqual(result, expected)


class Test_element_arrays_and_dims(tests.IrisTest):
    def test_single_field(self):
        field = _make_field(2013)
        collation = BasicFieldCollation([field])
        self.assertEqual(collation.element_arrays_and_dims, {})

    def test_t1(self):
        collation = BasicFieldCollation(
            [_make_field(lbyr=2013), _make_field(lbyr=2014)]
        )
        result = collation.element_arrays_and_dims
        self.assertEqual(list(result.keys()), ["t1"])
        values, dims = result["t1"]
        self.assertArrayEqual(
            values, [datetime(2013, 1, 1), datetime(2014, 1, 1)]
        )
        self.assertEqual(dims, (0,))

    def test_t1_and_t2(self):
        collation = BasicFieldCollation(
            [
                _make_field(lbyr=2013, lbyrd=2000),
                _make_field(lbyr=2014, lbyrd=2001),
                _make_field(lbyr=2015, lbyrd=2002),
            ]
        )
        result = collation.element_arrays_and_dims
        self.assertEqual(set(result.keys()), set(["t1", "t2"]))
        values, dims = result["t1"]
        self.assertArrayEqual(
            values,
            [datetime(2013, 1, 1), datetime(2014, 1, 1), datetime(2015, 1, 1)],
        )
        self.assertEqual(dims, (0,))
        values, dims = result["t2"]
        self.assertArrayEqual(
            values,
            [datetime(2000, 1, 1), datetime(2001, 1, 1), datetime(2002, 1, 1)],
        )
        self.assertEqual(dims, (0,))

    def test_t1_and_t2_and_lbft(self):
        collation = BasicFieldCollation(
            [
                _make_field(lbyr=1, lbyrd=15, lbft=6),
                _make_field(lbyr=1, lbyrd=16, lbft=9),
                _make_field(lbyr=11, lbyrd=25, lbft=6),
                _make_field(lbyr=11, lbyrd=26, lbft=9),
            ]
        )
        result = collation.element_arrays_and_dims
        self.assertEqual(set(result.keys()), set(["t1", "t2", "lbft"]))
        values, dims = result["t1"]
        self.assertArrayEqual(values, [datetime(1, 1, 1), datetime(11, 1, 1)])
        self.assertEqual(dims, (0,))
        values, dims = result["t2"]
        self.assertArrayEqual(
            values,
            [
                [datetime(15, 1, 1), datetime(16, 1, 1)],
                [datetime(25, 1, 1), datetime(26, 1, 1)],
            ],
        )
        self.assertEqual(dims, (0, 1))
        values, dims = result["lbft"]
        self.assertArrayEqual(values, [6, 9])
        self.assertEqual(dims, (1,))

    def test_blev(self):
        collation = BasicFieldCollation(
            [_make_field(blev=1), _make_field(blev=2)]
        )
        result = collation.element_arrays_and_dims
        keys = set(
            ["blev", "brsvd1", "brsvd2", "brlev", "bhrlev", "lblev", "bhlev"]
        )
        self.assertEqual(set(result.keys()), keys)
        values, dims = result["blev"]
        self.assertArrayEqual(values, [1, 2])
        self.assertEqual(dims, (0,))

    def test_bhlev(self):
        collation = BasicFieldCollation(
            [_make_field(blev=0, bhlev=1), _make_field(blev=1, bhlev=2)]
        )
        result = collation.element_arrays_and_dims
        keys = set(
            ["blev", "brsvd1", "brsvd2", "brlev", "bhrlev", "lblev", "bhlev"]
        )
        self.assertEqual(set(result.keys()), keys)
        values, dims = result["bhlev"]
        self.assertArrayEqual(values, [1, 2])
        self.assertEqual(dims, (0,))


class Test__time_comparable_int(tests.IrisTest):
    def test(self):
        # Define a list of date-time tuples, which should remain both all
        # distinct and in ascending order when converted...
        test_date_tuples = [
            # Increment each component in turn to check that all are handled.
            (2004, 1, 1, 0, 0, 0),
            (2004, 1, 1, 0, 0, 1),
            (2004, 1, 1, 0, 1, 0),
            (2004, 1, 1, 1, 0, 0),
            (2004, 1, 2, 0, 0, 0),
            (2004, 2, 1, 0, 0, 0),
            # Go across 2004-02-29 leap-day, and on to "Feb 31 .. Mar 1".
            (2004, 2, 27, 0, 0, 0),
            (2004, 2, 28, 0, 0, 0),
            (2004, 2, 29, 0, 0, 0),
            (2004, 2, 30, 0, 0, 0),
            (2004, 2, 31, 0, 0, 0),
            (2004, 3, 1, 0, 0, 0),
            (2005, 1, 1, 0, 0, 0),
        ]

        collation = BasicFieldCollation(["foo", "bar"])
        test_date_ints = [
            collation._time_comparable_int(*test_tuple)
            for test_tuple in test_date_tuples
        ]
        # Check all values are distinct.
        self.assertEqual(len(test_date_ints), len(set(test_date_ints)))
        # Check all values are in order.
        self.assertEqual(test_date_ints, sorted(test_date_ints))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""
This file currently exists as a simple demonstration of Airspeed Velocity
performance testing. it is expected to be expanded/replaced in future.

"""

# import iris tests first so that some things can be initialised before
# importing anything else.
from iris import tests

import iris
from iris.analysis import AreaWeighted


@tests.skip_data
class RegriddingTests:
    def setup(self):
        # Prepare a cube and a regridding scheme.
        file_path = tests.get_data_path(
            ["NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc"]
        )
        self.cube = iris.load_cube(file_path)
        self.scheme_area_w = AreaWeighted()

    def time_regrid_area_w(self):
        # Regrid the cube onto itself.
        self.cube.regrid(self.cube, self.scheme_area_w)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the class
:class:`iris.fileformats.um._fast_load.FieldCollation`.

This only tests the additional functionality for recording file locations of
PPFields that make loaded cubes.
The original class is the baseclass of this, now renamed 'BasicFieldCollation'.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import numpy as np

import iris

from iris.tests.integration.fast_load.test_fast_load import Mixin_FieldTest


class TestFastCallbackLocationInfo(Mixin_FieldTest, tests.IrisTest):
    do_fast_loads = True

    def setUp(self):
        # Call parent setup.
        super().setUp()

        # Create a basic load test case.
        self.callback_collations = []
        self.callback_filepaths = []

        def fast_load_callback(cube, collation, filename):
            self.callback_collations.append(collation)
            self.callback_filepaths.append(filename)

        flds = self.fields(c_t="11112222", c_h="11221122", phn="01010101")
        self.test_filepath = self.save_fieldcubes(flds)
        iris.load(self.test_filepath, callback=fast_load_callback)

    def test_callback_collations_filepaths(self):
        self.assertEqual(len(self.callback_collations), 2)
        self.assertEqual(
            self.callback_collations[0].data_filepath, self.test_filepath
        )
        self.assertEqual(
            self.callback_collations[1].data_filepath, self.test_filepath
        )

    def test_callback_collations_field_indices(self):
        self.assertEqual(
            self.callback_collations[0].data_field_indices.dtype, np.int64
        )
        self.assertArrayEqual(
            self.callback_collations[0].data_field_indices, [[1, 3], [5, 7]]
        )

        self.assertEqual(
            self.callback_collations[1].data_field_indices.dtype, np.int64
        )
        self.assertArrayEqual(
            self.callback_collations[1].data_field_indices, [[0, 2], [4, 6]]
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coord_systems.LambertConformal` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.coord_systems import LambertConformal


class Test_init_defaults(tests.IrisTest):
    # NOTE: most of the testing for LambertConformal is in the legacy test
    # module 'iris.tests.test_coordsystem'.
    # This class *only* tests the defaults for optional constructor args.

    def test_set_optional_args(self):
        # Check that setting the optional (non-ellipse) args works.
        # (Except secant_latitudes, which are done separately).
        crs = LambertConformal(
            central_lat=25.3,
            central_lon=-172,
            false_easting=100,
            false_northing=-200,
        )
        self.assertEqualAndKind(crs.central_lat, 25.3)
        self.assertEqualAndKind(crs.central_lon, -172.0)
        self.assertEqualAndKind(crs.false_easting, 100.0)
        self.assertEqualAndKind(crs.false_northing, -200.0)

    def test_set_one_parallel(self):
        # Check that setting the optional (non-ellipse) args works.
        # (Except secant_latitudes, which are done separately).
        crs = LambertConformal(secant_latitudes=-44)
        self.assertEqual(len(crs.secant_latitudes), 1)
        self.assertEqualAndKind(crs.secant_latitudes[0], -44.0)

    def test_set_two_parallels(self):
        # Check that setting the optional (non-ellipse) args works.
        # (Except secant_latitudes, which are done separately).
        crs = LambertConformal(secant_latitudes=[43, -7])
        self.assertEqual(len(crs.secant_latitudes), 2)
        self.assertEqualAndKind(crs.secant_latitudes[0], 43.0)
        self.assertEqualAndKind(crs.secant_latitudes[1], -7.0)

    def _check_crs_defaults(self, crs):
        # Check for property defaults when no kwargs options were set.
        # NOTE: except ellipsoid, which is done elsewhere.
        self.assertEqualAndKind(crs.central_lat, 39.0)
        self.assertEqualAndKind(crs.central_lon, -96.0)
        self.assertEqualAndKind(crs.false_easting, 0.0)
        self.assertEqualAndKind(crs.false_northing, 0.0)
        self.assertEqual(len(crs.secant_latitudes), 2)
        self.assertEqualAndKind(crs.secant_latitudes[0], 33.0)
        self.assertEqualAndKind(crs.secant_latitudes[1], 45.0)

    def test_no_optional_args(self):
        # Check expected defaults with no optional args.
        crs = LambertConformal()
        self._check_crs_defaults(crs)

    def test_optional_args_None(self):
        # Check expected defaults with optional args=None.
        crs = LambertConformal(
            central_lat=None,
            central_lon=None,
            false_easting=None,
            false_northing=None,
            secant_latitudes=None,
        )
        self._check_crs_defaults(crs)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris.fileformats.netcdf._FillValueMaskCheckAndStoreTarget`
class.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.fileformats.netcdf import _FillValueMaskCheckAndStoreTarget


class Test__FillValueMaskCheckAndStoreTarget(tests.IrisTest):
    def _call_target(self, fill_value, keys, vals):
        inner_target = mock.MagicMock()
        target = _FillValueMaskCheckAndStoreTarget(
            inner_target, fill_value=fill_value
        )

        for key, val in zip(keys, vals):
            target[key] = val

        calls = [mock.call(key, val) for key, val in zip(keys, vals)]
        inner_target.__setitem__.assert_has_calls(calls)

        return target

    def test___setitem__(self):
        self._call_target(None, [1], [2])

    def test_no_fill_value_not_masked(self):
        # Test when the fill value is not present and the data is not masked
        keys = [slice(0, 10), slice(10, 15)]
        vals = [np.arange(10), np.arange(5)]
        fill_value = 16
        target = self._call_target(fill_value, keys, vals)
        self.assertFalse(target.contains_value)
        self.assertFalse(target.is_masked)

    def test_contains_fill_value_not_masked(self):
        # Test when the fill value is present and the data is not masked
        keys = [slice(0, 10), slice(10, 15)]
        vals = [np.arange(10), np.arange(5)]
        fill_value = 5
        target = self._call_target(fill_value, keys, vals)
        self.assertTrue(target.contains_value)
        self.assertFalse(target.is_masked)

    def test_no_fill_value_masked(self):
        # Test when the fill value is not present and the data is masked
        keys = [slice(0, 10), slice(10, 15)]
        vals = [np.arange(10), np.ma.masked_equal(np.arange(5), 3)]
        fill_value = 16
        target = self._call_target(fill_value, keys, vals)
        self.assertFalse(target.contains_value)
        self.assertTrue(target.is_masked)

    def test_contains_fill_value_masked(self):
        # Test when the fill value is present and the data is masked
        keys = [slice(0, 10), slice(10, 15)]
        vals = [np.arange(10), np.ma.masked_equal(np.arange(5), 3)]
        fill_value = 5
        target = self._call_target(fill_value, keys, vals)
        self.assertTrue(target.contains_value)
        self.assertTrue(target.is_masked)

    def test_fill_value_None(self):
        # Test when the fill value is None
        keys = [slice(0, 10), slice(10, 15)]
        vals = [np.arange(10), np.arange(5)]
        fill_value = None
        target = self._call_target(fill_value, keys, vals)
        self.assertFalse(target.contains_value)

    def test_contains_masked_fill_value(self):
        # Test when the fill value is present but masked the data is masked
        keys = [slice(0, 10), slice(10, 15)]
        vals = [np.arange(10), np.ma.masked_equal(np.arange(10, 15), 13)]
        fill_value = 13
        target = self._call_target(fill_value, keys, vals)
        self.assertFalse(target.contains_value)
        self.assertTrue(target.is_masked)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot._fixup_dates` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from cf_units import Unit
import datetime
import cftime

from iris.coords import AuxCoord
from iris.plot import _fixup_dates


class Test(tests.IrisTest):
    def test_gregorian_calendar(self):
        unit = Unit("hours since 2000-04-13 00:00:00", calendar="gregorian")
        coord = AuxCoord([1, 3, 6], "time", units=unit)
        result = _fixup_dates(coord, coord.points)
        expected = [
            datetime.datetime(2000, 4, 13, 1),
            datetime.datetime(2000, 4, 13, 3),
            datetime.datetime(2000, 4, 13, 6),
        ]
        self.assertArrayEqual(result, expected)

    def test_gregorian_calendar_sub_second(self):
        unit = Unit("seconds since 2000-04-13 00:00:00", calendar="gregorian")
        coord = AuxCoord([1, 1.25, 1.5], "time", units=unit)
        result = _fixup_dates(coord, coord.points)
        expected = [
            datetime.datetime(2000, 4, 13, 0, 0, 1),
            datetime.datetime(2000, 4, 13, 0, 0, 1),
            datetime.datetime(2000, 4, 13, 0, 0, 2),
        ]
        self.assertArrayEqual(result, expected)

    @tests.skip_nc_time_axis
    def test_360_day_calendar(self):
        unit = Unit("days since 2000-02-25 00:00:00", calendar="360_day")
        coord = AuxCoord([3, 4, 5], "time", units=unit)
        result = _fixup_dates(coord, coord.points)
        expected_datetimes = [
            cftime.datetime(2000, 2, 28),
            cftime.datetime(2000, 2, 29),
            cftime.datetime(2000, 2, 30),
        ]
        self.assertArrayEqual(
            [cdt.datetime for cdt in result], expected_datetimes
        )

    @tests.skip_nc_time_axis
    def test_365_day_calendar(self):
        unit = Unit("minutes since 2000-02-25 00:00:00", calendar="365_day")
        coord = AuxCoord([30, 60, 150], "time", units=unit)
        result = _fixup_dates(coord, coord.points)
        expected_datetimes = [
            cftime.datetime(2000, 2, 25, 0, 30),
            cftime.datetime(2000, 2, 25, 1, 0),
            cftime.datetime(2000, 2, 25, 2, 30),
        ]
        self.assertArrayEqual(
            [cdt.datetime for cdt in result], expected_datetimes
        )

    @tests.skip_nc_time_axis
    def test_360_day_calendar_attribute(self):
        unit = Unit("days since 2000-02-01 00:00:00", calendar="360_day")
        coord = AuxCoord([0, 3, 6], "time", units=unit)
        result = _fixup_dates(coord, coord.points)
        self.assertEqual(result[0].calendar, "360_day")


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris._lazy data.map_complete_blocks`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import unittest
import numpy as np
import dask.array as da

from iris._lazy_data import is_lazy_data, map_complete_blocks


def create_mock_cube(array):
    cube = unittest.mock.Mock()
    cube_data = unittest.mock.PropertyMock(return_value=array)
    type(cube).data = cube_data
    cube.dtype = array.dtype
    cube.has_lazy_data = unittest.mock.Mock(return_value=is_lazy_data(array))
    cube.lazy_data = unittest.mock.Mock(return_value=array)
    cube.shape = array.shape
    return cube, cube_data


class Test_map_complete_blocks(tests.IrisTest):
    def setUp(self):
        self.array = np.arange(8).reshape(2, 4)
        self.func = lambda chunk: chunk + 1
        self.func_result = self.array + 1

    def test_non_lazy_input(self):
        # Check that a non-lazy input doesn't trip up the functionality.
        cube, cube_data = create_mock_cube(self.array)
        result = map_complete_blocks(
            cube, self.func, dims=(1,), out_sizes=(4,)
        )
        self.assertFalse(is_lazy_data(result))
        self.assertArrayEqual(result, self.func_result)
        # check correct data was accessed
        cube.lazy_data.assert_not_called()
        cube_data.assert_called_once()

    def test_lazy_input(self):
        lazy_array = da.asarray(self.array, chunks=((1, 1), (4,)))
        cube, cube_data = create_mock_cube(lazy_array)
        result = map_complete_blocks(
            cube, self.func, dims=(1,), out_sizes=(4,)
        )
        self.assertTrue(is_lazy_data(result))
        self.assertArrayEqual(result.compute(), self.func_result)
        # check correct data was accessed
        cube.lazy_data.assert_called_once()
        cube_data.assert_not_called()

    def test_rechunk(self):
        lazy_array = da.asarray(self.array, chunks=((1, 1), (2, 2)))
        cube, _ = create_mock_cube(lazy_array)
        result = map_complete_blocks(
            cube, self.func, dims=(1,), out_sizes=(4,)
        )
        self.assertTrue(is_lazy_data(result))
        self.assertArrayEqual(result.compute(), self.func_result)

    def test_different_out_shape(self):
        lazy_array = da.asarray(self.array, chunks=((1, 1), (4,)))
        cube, _ = create_mock_cube(lazy_array)

        def func(_):
            return np.arange(2).reshape(1, 2)

        func_result = [[0, 1], [0, 1]]
        result = map_complete_blocks(cube, func, dims=(1,), out_sizes=(2,))
        self.assertTrue(is_lazy_data(result))
        self.assertArrayEqual(result.compute(), func_result)

    def test_multidimensional_input(self):
        array = np.arange(2 * 3 * 4).reshape(2, 3, 4)
        lazy_array = da.asarray(array, chunks=((1, 1), (1, 2), (4,)))
        cube, _ = create_mock_cube(lazy_array)
        result = map_complete_blocks(
            cube, self.func, dims=(1, 2), out_sizes=(3, 4)
        )
        self.assertTrue(is_lazy_data(result))
        self.assertArrayEqual(result.compute(), array + 1)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the cube merging mechanism.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

from collections.abc import Iterable
import datetime
import itertools
import numpy as np
import numpy.ma as ma

import iris
from iris._lazy_data import as_lazy_data
import iris.cube
from iris.coords import DimCoord, AuxCoord
import iris.exceptions
import iris.tests.stock


class TestMixin:
    """
    Mix-in class for attributes & utilities common to these test cases.

    """

    def test_normal_cubes(self):
        cubes = iris.load(self._data_path)
        self.assertEqual(len(cubes), self._num_cubes)
        self.assertCML(cubes, ["merge", self._prefix + ".cml"])

    def test_remerge(self):
        # After the merge process the coordinates within each cube can be in a
        # different order. Until that changes we can't compare the cubes
        # directly or with the CML ... so we just make sure the count stays
        # the same.
        cubes = iris.load(self._data_path)
        cubes2 = cubes.merge()
        self.assertEqual(len(cubes), len(cubes2))

    def test_duplication(self):
        cubes = iris.load(self._data_path)
        self.assertRaises(
            iris.exceptions.DuplicateDataError, (cubes + cubes).merge
        )
        cubes2 = (cubes + cubes).merge(unique=False)
        self.assertEqual(len(cubes2), 2 * len(cubes))


@tests.skip_data
class TestSingleCube(tests.IrisTest, TestMixin):
    def setUp(self):
        self._data_path = tests.get_data_path(("PP", "globClim1", "theta.pp"))
        self._num_cubes = 1
        self._prefix = "theta"


@tests.skip_data
class TestMultiCube(tests.IrisTest, TestMixin):
    def setUp(self):
        self._data_path = tests.get_data_path(
            ("PP", "globClim1", "dec_subset.pp")
        )
        self._num_cubes = 4
        self._prefix = "dec"

    def test_coord_attributes(self):
        def custom_coord_callback(cube, field, filename):
            cube.coord("time").attributes["monty"] = "python"
            cube.coord("time").attributes["brain"] = "hurts"

        # Load slices, decorating a coord with custom attributes
        cubes = iris.load_raw(self._data_path, callback=custom_coord_callback)
        # Merge
        merged = iris.cube.CubeList(cubes).merge()
        # Check the custom attributes are in the merged cube
        for cube in merged:
            assert cube.coord("time").attributes["monty"] == "python"
            assert cube.coord("time").attributes["brain"] == "hurts"


@tests.skip_data
class TestColpex(tests.IrisTest):
    def setUp(self):
        self._data_path = tests.get_data_path(
            ("PP", "COLPEX", "small_colpex_theta_p_alt.pp")
        )

    def test_colpex(self):
        cubes = iris.load(self._data_path)
        self.assertEqual(len(cubes), 3)
        self.assertCML(cubes, ("COLPEX", "small_colpex_theta_p_alt.cml"))


class TestDataMergeCombos(tests.IrisTest):
    def _make_data(
        self,
        data,
        dtype=np.dtype("int32"),
        fill_value=None,
        mask=None,
        lazy=False,
        N=3,
    ):
        if isinstance(data, Iterable):
            shape = (len(data), N, N)
            data = np.array(data).reshape(-1, 1, 1)
        else:
            shape = (N, N)
        if mask is not None:
            payload = ma.empty(shape, dtype=dtype, fill_value=fill_value)
            payload.data[:] = data
            if isinstance(mask, bool):
                payload.mask = mask
            else:
                payload[mask] = ma.masked
        else:
            payload = np.empty(shape, dtype=dtype)
            payload[:] = data
        if lazy:
            payload = as_lazy_data(payload)
        return payload

    def _make_cube(
        self,
        data,
        dtype=np.dtype("int32"),
        fill_value=None,
        mask=None,
        lazy=False,
        N=3,
    ):
        x = np.arange(N)
        y = np.arange(N)
        payload = self._make_data(
            data, dtype=dtype, fill_value=fill_value, mask=mask, lazy=lazy, N=N
        )
        cube = iris.cube.Cube(payload)
        lat = DimCoord(y, standard_name="latitude", units="degrees")
        cube.add_dim_coord(lat, 0)
        lon = DimCoord(x, standard_name="longitude", units="degrees")
        cube.add_dim_coord(lon, 1)
        height = DimCoord(data, standard_name="height", units="m")
        cube.add_aux_coord(height)
        return cube

    @staticmethod
    def _expected_fill_value(fill0="none", fill1="none"):
        result = None
        if fill0 != "none" or fill1 != "none":
            if fill0 == "none":
                result = fill1
            elif fill1 == "none":
                result = fill0
            elif fill0 == fill1:
                result = fill0
        return result

    def _check_fill_value(self, result, fill0="none", fill1="none"):
        expected_fill_value = self._expected_fill_value(fill0, fill1)
        if expected_fill_value is None:
            data = result.data
            if ma.isMaskedArray(data):
                np_fill_value = ma.masked_array(
                    0, dtype=result.dtype
                ).fill_value
                self.assertEqual(data.fill_value, np_fill_value)
        else:
            data = result.data
            if ma.isMaskedArray(data):
                self.assertEqual(data.fill_value, expected_fill_value)

    def setUp(self):
        self.dtype = np.dtype("int32")
        fill_value = 1234
        self.lazy_combos = itertools.product([False, True], [False, True])
        fill_combos = itertools.product([None, fill_value], [fill_value, None])
        single_fill_combos = itertools.product([None, fill_value])
        self.combos = itertools.product(self.lazy_combos, fill_combos)
        self.mixed_combos = itertools.product(
            self.lazy_combos, single_fill_combos
        )

    def test__ndarray_ndarray(self):
        for (lazy0, lazy1) in self.lazy_combos:
            cubes = iris.cube.CubeList()
            cubes.append(self._make_cube(0, dtype=self.dtype, lazy=lazy0))
            cubes.append(self._make_cube(1, dtype=self.dtype, lazy=lazy1))
            result = cubes.merge_cube()
            expected = self._make_data([0, 1], dtype=self.dtype)
            self.assertArrayEqual(result.data, expected)
            self.assertEqual(result.dtype, self.dtype)
            self._check_fill_value(result)

    def test__masked_masked(self):
        for (lazy0, lazy1), (fill0, fill1) in self.combos:
            cubes = iris.cube.CubeList()
            mask = ((0,), (0,))
            cubes.append(
                self._make_cube(
                    0,
                    mask=mask,
                    lazy=lazy0,
                    dtype=self.dtype,
                    fill_value=fill0,
                )
            )
            mask = ((1,), (1,))
            cubes.append(
                self._make_cube(
                    1,
                    mask=mask,
                    lazy=lazy1,
                    dtype=self.dtype,
                    fill_value=fill1,
                )
            )
            result = cubes.merge_cube()
            mask = ((0, 1), (0, 1), (0, 1))
            expected_fill_value = self._expected_fill_value(fill0, fill1)
            expected = self._make_data(
                [0, 1],
                mask=mask,
                dtype=self.dtype,
                fill_value=expected_fill_value,
            )
            self.assertMaskedArrayEqual(result.data, expected)
            self.assertEqual(result.dtype, self.dtype)
            self._check_fill_value(result, fill0, fill1)

    def test__ndarray_masked(self):
        for (lazy0, lazy1), (fill,) in self.mixed_combos:
            cubes = iris.cube.CubeList()
            cubes.append(self._make_cube(0, lazy=lazy0, dtype=self.dtype))
            mask = [(0, 1), (0, 1)]
            cubes.append(
                self._make_cube(
                    1, mask=mask, lazy=lazy1, dtype=self.dtype, fill_value=fill
                )
            )
            result = cubes.merge_cube()
            mask = [(1, 1), (0, 1), (0, 1)]
            expected_fill_value = self._expected_fill_value(fill)
            expected = self._make_data(
                [0, 1],
                mask=mask,
                dtype=self.dtype,
                fill_value=expected_fill_value,
            )
            self.assertMaskedArrayEqual(result.data, expected)
            self.assertEqual(result.dtype, self.dtype)
            self._check_fill_value(result, fill1=fill)

    def test__masked_ndarray(self):
        for (lazy0, lazy1), (fill,) in self.mixed_combos:
            cubes = iris.cube.CubeList()
            mask = [(0, 1), (0, 1)]
            cubes.append(
                self._make_cube(
                    0, mask=mask, lazy=lazy0, dtype=self.dtype, fill_value=fill
                )
            )
            cubes.append(self._make_cube(1, lazy=lazy1, dtype=self.dtype))
            result = cubes.merge_cube()
            mask = [(0, 0), (0, 1), (0, 1)]
            expected_fill_value = self._expected_fill_value(fill)
            expected = self._make_data(
                [0, 1],
                mask=mask,
                dtype=self.dtype,
                fill_value=expected_fill_value,
            )
            self.assertMaskedArrayEqual(result.data, expected)
            self.assertEqual(result.dtype, self.dtype)
            self._check_fill_value(result, fill0=fill)

    def test_maksed_array_preserved(self):
        for (lazy0, lazy1), (fill,) in self.mixed_combos:
            cubes = iris.cube.CubeList()
            mask = False
            cubes.append(
                self._make_cube(
                    0, mask=mask, lazy=lazy0, dtype=self.dtype, fill_value=fill
                )
            )
            cubes.append(self._make_cube(1, lazy=lazy1, dtype=self.dtype))
            result = cubes.merge_cube()
            mask = False
            expected_fill_value = self._expected_fill_value(fill)
            expected = self._make_data(
                [0, 1],
                mask=mask,
                dtype=self.dtype,
                fill_value=expected_fill_value,
            )
            self.assertEqual(type(result.data), ma.MaskedArray)
            self.assertMaskedArrayEqual(result.data, expected)
            self.assertEqual(result.dtype, self.dtype)
            self._check_fill_value(result, fill0=fill)

    def test_fill_value_invariant_to_order__same_non_None(self):
        fill_value = 1234
        cubes = [
            self._make_cube(i, mask=True, fill_value=fill_value)
            for i in range(3)
        ]
        for combo in itertools.permutations(cubes):
            result = iris.cube.CubeList(combo).merge_cube()
            self.assertEqual(result.data.fill_value, fill_value)

    def test_fill_value_invariant_to_order__all_None(self):
        cubes = [
            self._make_cube(i, mask=True, fill_value=None) for i in range(3)
        ]
        for combo in itertools.permutations(cubes):
            result = iris.cube.CubeList(combo).merge_cube()
            np_fill_value = ma.masked_array(0, dtype=result.dtype).fill_value
            self.assertEqual(result.data.fill_value, np_fill_value)

    def test_fill_value_invariant_to_order__different_non_None(self):
        cubes = [self._make_cube(0, mask=True, fill_value=1234)]
        cubes.append(self._make_cube(1, mask=True, fill_value=2341))
        cubes.append(self._make_cube(2, mask=True, fill_value=3412))
        cubes.append(self._make_cube(3, mask=True, fill_value=4123))
        for combo in itertools.permutations(cubes):
            result = iris.cube.CubeList(combo).merge_cube()
            np_fill_value = ma.masked_array(0, dtype=result.dtype).fill_value
            self.assertEqual(result.data.fill_value, np_fill_value)

    def test_fill_value_invariant_to_order__mixed(self):
        cubes = [self._make_cube(0, mask=True, fill_value=None)]
        cubes.append(self._make_cube(1, mask=True, fill_value=1234))
        cubes.append(self._make_cube(2, mask=True, fill_value=4321))
        for combo in itertools.permutations(cubes):
            result = iris.cube.CubeList(combo).merge_cube()
            np_fill_value = ma.masked_array(0, dtype=result.dtype).fill_value
            self.assertEqual(result.data.fill_value, np_fill_value)


@tests.skip_data
class TestDataMerge(tests.IrisTest):
    def test_extended_proxy_data(self):
        # Get the empty theta cubes for T+1.5 and T+2
        data_path = tests.get_data_path(
            ("PP", "COLPEX", "theta_and_orog_subset.pp")
        )
        phenom_constraint = iris.Constraint("air_potential_temperature")
        datetime_1 = datetime.datetime(2009, 9, 9, 17, 20)
        datetime_2 = datetime.datetime(2009, 9, 9, 17, 50)
        time_constraint1 = iris.Constraint(time=datetime_1)
        time_constraint2 = iris.Constraint(time=datetime_2)
        time_constraint_1_and_2 = iris.Constraint(
            time=lambda c: c in (datetime_1, datetime_2)
        )
        cube1 = iris.load_cube(data_path, phenom_constraint & time_constraint1)
        cube2 = iris.load_cube(data_path, phenom_constraint & time_constraint2)

        # Merge the two halves
        cubes = iris.cube.CubeList([cube1, cube2]).merge(True)
        self.assertCML(cubes, ("merge", "theta_two_times.cml"))

        # Make sure we get the same result directly from load
        cubes = iris.load_cube(
            data_path, phenom_constraint & time_constraint_1_and_2
        )
        self.assertCML(cubes, ("merge", "theta_two_times.cml"))

    def test_real_data(self):
        data_path = tests.get_data_path(("PP", "globClim1", "theta.pp"))
        cubes = iris.load_raw(data_path)
        # Force the source 2-D cubes to load their data before the merge
        for cube in cubes:
            _ = cube.data
        cubes = cubes.merge()
        self.assertCML(cubes, ["merge", "theta.cml"])


class TestDimensionSplitting(tests.IrisTest):
    def _make_cube(self, a, b, c, data):
        cube_data = np.empty((4, 5), dtype=np.float32)
        cube_data[:] = data
        cube = iris.cube.Cube(cube_data)
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3, 4], dtype=np.int32),
                long_name="x",
                units="1",
            ),
            1,
        )
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3], dtype=np.int32),
                long_name="y",
                units="1",
            ),
            0,
        )
        cube.add_aux_coord(
            DimCoord(np.array([a], dtype=np.int32), long_name="a", units="1")
        )
        cube.add_aux_coord(
            DimCoord(np.array([b], dtype=np.int32), long_name="b", units="1")
        )
        cube.add_aux_coord(
            DimCoord(np.array([c], dtype=np.int32), long_name="c", units="1")
        )
        return cube

    def test_single_split(self):
        # Test what happens when a cube forces a simple, two-way split.
        cubes = []
        cubes.append(self._make_cube(0, 0, 0, 0))
        cubes.append(self._make_cube(0, 1, 1, 1))
        cubes.append(self._make_cube(1, 0, 2, 2))
        cubes.append(self._make_cube(1, 1, 3, 3))
        cubes.append(self._make_cube(2, 0, 4, 4))
        cubes.append(self._make_cube(2, 1, 5, 5))
        cube = iris.cube.CubeList(cubes).merge()
        self.assertCML(cube, ("merge", "single_split.cml"))

    def test_multi_split(self):
        # Test what happens when a cube forces a three-way split.
        cubes = []
        cubes.append(self._make_cube(0, 0, 0, 0))
        cubes.append(self._make_cube(0, 0, 1, 1))
        cubes.append(self._make_cube(0, 1, 0, 2))
        cubes.append(self._make_cube(0, 1, 1, 3))
        cubes.append(self._make_cube(1, 0, 0, 4))
        cubes.append(self._make_cube(1, 0, 1, 5))
        cubes.append(self._make_cube(1, 1, 0, 6))
        cubes.append(self._make_cube(1, 1, 1, 7))
        cubes.append(self._make_cube(2, 0, 0, 8))
        cubes.append(self._make_cube(2, 0, 1, 9))
        cubes.append(self._make_cube(2, 1, 0, 10))
        cubes.append(self._make_cube(2, 1, 1, 11))
        cube = iris.cube.CubeList(cubes).merge()
        self.assertCML(cube, ("merge", "multi_split.cml"))


class TestCombination(tests.IrisTest):
    def _make_cube(self, a, b, c, d, data=0):
        cube_data = np.empty((4, 5), dtype=np.float32)
        cube_data[:] = data
        cube = iris.cube.Cube(cube_data)
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3, 4], dtype=np.int32),
                long_name="x",
                units="1",
            ),
            1,
        )
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3], dtype=np.int32),
                long_name="y",
                units="1",
            ),
            0,
        )

        for name, value in zip(["a", "b", "c", "d"], [a, b, c, d]):
            dtype = np.str if isinstance(value, str) else np.float32
            cube.add_aux_coord(
                AuxCoord(
                    np.array([value], dtype=dtype), long_name=name, units="1"
                )
            )

        return cube

    def test_separable_combination(self):
        cubes = iris.cube.CubeList()
        cubes.append(
            self._make_cube(
                "2005", "ECMWF", "HOPE-E, Sys 1, Met 1, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2005", "ECMWF", "HOPE-E, Sys 1, Met 1, ENSEMBLES", 1
            )
        )
        cubes.append(
            self._make_cube(
                "2005", "ECMWF", "HOPE-E, Sys 1, Met 1, ENSEMBLES", 2
            )
        )
        cubes.append(
            self._make_cube(
                "2026", "UK Met Office", "HadGEM2, Sys 1, Met 1, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2026", "UK Met Office", "HadGEM2, Sys 1, Met 1, ENSEMBLES", 1
            )
        )
        cubes.append(
            self._make_cube(
                "2026", "UK Met Office", "HadGEM2, Sys 1, Met 1, ENSEMBLES", 2
            )
        )
        cubes.append(
            self._make_cube(
                "2002", "CERFACS", "GELATO, Sys 0, Met 1, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2002", "CERFACS", "GELATO, Sys 0, Met 1, ENSEMBLES", 1
            )
        )
        cubes.append(
            self._make_cube(
                "2002", "CERFACS", "GELATO, Sys 0, Met 1, ENSEMBLES", 2
            )
        )
        cubes.append(
            self._make_cube(
                "2002", "IFM-GEOMAR", "ECHAM5, Sys 1, Met 10, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2002", "IFM-GEOMAR", "ECHAM5, Sys 1, Met 10, ENSEMBLES", 1
            )
        )
        cubes.append(
            self._make_cube(
                "2002", "IFM-GEOMAR", "ECHAM5, Sys 1, Met 10, ENSEMBLES", 2
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 10, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 11, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 12, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 13, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 14, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 15, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 16, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 17, ENSEMBLES", 0
            )
        )
        cubes.append(
            self._make_cube(
                "2502", "UK Met Office", "HadCM3, Sys 51, Met 18, ENSEMBLES", 0
            )
        )
        cube = cubes.merge()
        self.assertCML(
            cube, ("merge", "separable_combination.cml"), checksum=False
        )


class TestDimSelection(tests.IrisTest):
    def _make_cube(self, a, b, data=0, a_dim=False, b_dim=False):
        cube_data = np.empty((4, 5), dtype=np.float32)
        cube_data[:] = data
        cube = iris.cube.Cube(cube_data)
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3, 4], dtype=np.int32),
                long_name="x",
                units="1",
            ),
            1,
        )
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3], dtype=np.int32),
                long_name="y",
                units="1",
            ),
            0,
        )

        for name, value, dim in zip(["a", "b"], [a, b], [a_dim, b_dim]):
            dtype = np.str if isinstance(value, str) else np.float32
            ctype = DimCoord if dim else AuxCoord
            coord = ctype(
                np.array([value], dtype=dtype), long_name=name, units="1"
            )
            cube.add_aux_coord(coord)

        return cube

    def test_string_a_with_aux(self):
        templates = (("a", 0), ("b", 1), ("c", 2), ("d", 3))
        cubes = [self._make_cube(a, b) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(
            cube, ("merge", "string_a_with_aux.cml"), checksum=False
        )
        self.assertIsInstance(cube.coord("a"), AuxCoord)
        self.assertIsInstance(cube.coord("b"), DimCoord)
        self.assertTrue(cube.coord("b") in cube.dim_coords)

    def test_string_b_with_aux(self):
        templates = ((0, "a"), (1, "b"), (2, "c"), (3, "d"))
        cubes = [self._make_cube(a, b) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(
            cube, ("merge", "string_b_with_aux.cml"), checksum=False
        )
        self.assertIsInstance(cube.coord("a"), DimCoord)
        self.assertTrue(cube.coord("a") in cube.dim_coords)
        self.assertIsInstance(cube.coord("b"), AuxCoord)

    def test_string_a_with_dim(self):
        templates = (("a", 0), ("b", 1), ("c", 2), ("d", 3))
        cubes = [self._make_cube(a, b, b_dim=True) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(
            cube, ("merge", "string_a_with_dim.cml"), checksum=False
        )
        self.assertIsInstance(cube.coord("a"), AuxCoord)
        self.assertIsInstance(cube.coord("b"), DimCoord)
        self.assertTrue(cube.coord("b") in cube.dim_coords)

    def test_string_b_with_dim(self):
        templates = ((0, "a"), (1, "b"), (2, "c"), (3, "d"))
        cubes = [self._make_cube(a, b, a_dim=True) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(
            cube, ("merge", "string_b_with_dim.cml"), checksum=False
        )
        self.assertIsInstance(cube.coord("a"), DimCoord)
        self.assertTrue(cube.coord("a") in cube.dim_coords)
        self.assertIsInstance(cube.coord("b"), AuxCoord)

    def test_string_a_b(self):
        templates = (("a", "0"), ("b", "1"), ("c", "2"), ("d", "3"))
        cubes = [self._make_cube(a, b) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(cube, ("merge", "string_a_b.cml"), checksum=False)
        self.assertIsInstance(cube.coord("a"), AuxCoord)
        self.assertIsInstance(cube.coord("b"), AuxCoord)

    def test_a_aux_b_aux(self):
        templates = ((0, 10), (1, 11), (2, 12), (3, 13))
        cubes = [self._make_cube(a, b) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(cube, ("merge", "a_aux_b_aux.cml"), checksum=False)
        self.assertIsInstance(cube.coord("a"), DimCoord)
        self.assertTrue(cube.coord("a") in cube.dim_coords)
        self.assertIsInstance(cube.coord("b"), DimCoord)
        self.assertTrue(cube.coord("b") in cube.aux_coords)

    def test_a_aux_b_dim(self):
        templates = ((0, 10), (1, 11), (2, 12), (3, 13))
        cubes = [self._make_cube(a, b, b_dim=True) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(cube, ("merge", "a_aux_b_dim.cml"), checksum=False)
        self.assertIsInstance(cube.coord("a"), DimCoord)
        self.assertTrue(cube.coord("a") in cube.aux_coords)
        self.assertIsInstance(cube.coord("b"), DimCoord)
        self.assertTrue(cube.coord("b") in cube.dim_coords)

    def test_a_dim_b_aux(self):
        templates = ((0, 10), (1, 11), (2, 12), (3, 13))
        cubes = [self._make_cube(a, b, a_dim=True) for a, b in templates]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(cube, ("merge", "a_dim_b_aux.cml"), checksum=False)
        self.assertIsInstance(cube.coord("a"), DimCoord)
        self.assertTrue(cube.coord("a") in cube.dim_coords)
        self.assertIsInstance(cube.coord("b"), DimCoord)
        self.assertTrue(cube.coord("b") in cube.aux_coords)

    def test_a_dim_b_dim(self):
        templates = ((0, 10), (1, 11), (2, 12), (3, 13))
        cubes = [
            self._make_cube(a, b, a_dim=True, b_dim=True) for a, b in templates
        ]
        cube = iris.cube.CubeList(cubes).merge()[0]
        self.assertCML(cube, ("merge", "a_dim_b_dim.cml"), checksum=False)
        self.assertIsInstance(cube.coord("a"), DimCoord)
        self.assertTrue(cube.coord("a") in cube.dim_coords)
        self.assertIsInstance(cube.coord("b"), DimCoord)
        self.assertTrue(cube.coord("b") in cube.aux_coords)


class TestTimeTripleMerging(tests.IrisTest):
    def _make_cube(self, a, b, c, data=0):
        cube_data = np.empty((4, 5), dtype=np.float32)
        cube_data[:] = data
        cube = iris.cube.Cube(cube_data)
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3, 4], dtype=np.int32),
                long_name="x",
                units="1",
            ),
            1,
        )
        cube.add_dim_coord(
            DimCoord(
                np.array([0, 1, 2, 3], dtype=np.int32),
                long_name="y",
                units="1",
            ),
            0,
        )
        cube.add_aux_coord(
            DimCoord(
                np.array([a], dtype=np.int32),
                standard_name="forecast_period",
                units="1",
            )
        )
        cube.add_aux_coord(
            DimCoord(
                np.array([b], dtype=np.int32),
                standard_name="forecast_reference_time",
                units="1",
            )
        )
        cube.add_aux_coord(
            DimCoord(
                np.array([c], dtype=np.int32), standard_name="time", units="1"
            )
        )
        return cube

    def _test_triples(self, triples, filename):
        cubes = [self._make_cube(fp, rt, t) for fp, rt, t in triples]
        cube = iris.cube.CubeList(cubes).merge()
        self.assertCML(
            cube, ("merge", "time_triple_" + filename + ".cml"), checksum=False
        )

    def test_single_forecast(self):
        # A single forecast series (i.e. from a single reference time)
        # => fp, t: 4; rt: scalar
        triples = (
            (0, 10, 10),
            (1, 10, 11),
            (2, 10, 12),
            (3, 10, 13),
        )
        self._test_triples(triples, "single_forecast")

    def test_successive_forecasts(self):
        # Three forecast series from successively later reference times
        # => rt, t: 3; fp, t: 4
        triples = (
            (0, 10, 10),
            (1, 10, 11),
            (2, 10, 12),
            (3, 10, 13),
            (0, 11, 11),
            (1, 11, 12),
            (2, 11, 13),
            (3, 11, 14),
            (0, 12, 12),
            (1, 12, 13),
            (2, 12, 14),
            (3, 12, 15),
        )
        self._test_triples(triples, "successive_forecasts")

    def test_time_vs_ref_time(self):
        # => fp, t: 4; fp, rt: 3
        triples = (
            (2, 10, 12),
            (3, 10, 13),
            (4, 10, 14),
            (5, 10, 15),
            (1, 11, 12),
            (2, 11, 13),
            (3, 11, 14),
            (4, 11, 15),
            (0, 12, 12),
            (1, 12, 13),
            (2, 12, 14),
            (3, 12, 15),
        )
        self._test_triples(triples, "time_vs_ref_time")

    def test_time_vs_forecast(self):
        # => rt, t: 4, fp, rt: 3
        triples = (
            (0, 10, 10),
            (0, 11, 11),
            (0, 12, 12),
            (0, 13, 13),
            (1, 9, 10),
            (1, 10, 11),
            (1, 11, 12),
            (1, 12, 13),
            (2, 8, 10),
            (2, 9, 11),
            (2, 10, 12),
            (2, 11, 13),
        )
        self._test_triples(triples, "time_vs_forecast")

    def test_time_non_dim_coord(self):
        # => rt: 1 fp, t (bounded): 2
        triples = (
            (5, 0, 2.5),
            (10, 0, 5),
        )
        cubes = [self._make_cube(fp, rt, t) for fp, rt, t in triples]
        for end_time, cube in zip([5, 10], cubes):
            cube.coord("time").bounds = [0, end_time]
        (cube,) = iris.cube.CubeList(cubes).merge()
        self.assertCML(
            cube,
            ("merge", "time_triple_time_non_dim_coord.cml"),
            checksum=False,
        )
        # make sure that forecast_period is the dimensioned coordinate (as time becomes an AuxCoord)
        self.assertEqual(
            cube.coord(dimensions=0, dim_coords=True).name(), "forecast_period"
        )

    def test_independent(self):
        # => fp: 2; rt: 2; t: 2
        triples = (
            (0, 10, 10),
            (0, 11, 10),
            (0, 10, 11),
            (0, 11, 11),
            (1, 10, 10),
            (1, 11, 10),
            (1, 10, 11),
            (1, 11, 11),
        )
        self._test_triples(triples, "independent")

    def test_series(self):
        # => fp, rt, t: 5 (with only t being definitive).
        triples = (
            (0, 10, 10),
            (0, 11, 11),
            (0, 12, 12),
            (1, 12, 13),
            (2, 12, 14),
        )
        self._test_triples(triples, "series")

    def test_non_expanding_dimension(self):
        triples = (
            (0, 10, 0),
            (0, 20, 1),
            (0, 20, 0),
        )
        # => fp: scalar; rt, t: 3 (with no time being definitive)
        self._test_triples(triples, "non_expanding")

    def test_duplicate_data(self):
        # test what happens when we have repeated time coordinates (i.e. duplicate data)
        cube1 = self._make_cube(0, 10, 0)
        cube2 = self._make_cube(1, 20, 1)
        cube3 = self._make_cube(1, 20, 1)

        # check that we get a duplicate data error when unique is True
        with self.assertRaises(iris.exceptions.DuplicateDataError):
            iris.cube.CubeList([cube1, cube2, cube3]).merge()

        cubes = iris.cube.CubeList([cube1, cube2, cube3]).merge(unique=False)
        self.assertCML(
            cubes, ("merge", "time_triple_duplicate_data.cml"), checksum=False
        )

    def test_simple1(self):
        cube1 = self._make_cube(0, 10, 0)
        cube2 = self._make_cube(1, 20, 1)
        cube3 = self._make_cube(2, 20, 0)
        cube = iris.cube.CubeList([cube1, cube2, cube3]).merge()
        self.assertCML(
            cube, ("merge", "time_triple_merging1.cml"), checksum=False
        )

    def test_simple2(self):
        cubes = iris.cube.CubeList(
            [
                self._make_cube(0, 0, 0),
                self._make_cube(1, 0, 1),
                self._make_cube(2, 0, 2),
                self._make_cube(0, 1, 3),
                self._make_cube(1, 1, 4),
                self._make_cube(2, 1, 5),
            ]
        )
        cube = cubes.merge()[0]
        self.assertCML(
            cube, ("merge", "time_triple_merging2.cml"), checksum=False
        )

        cube = iris.cube.CubeList(cubes[:-1]).merge()[0]
        self.assertCML(
            cube, ("merge", "time_triple_merging3.cml"), checksum=False
        )

    def test_simple3(self):
        cubes = iris.cube.CubeList(
            [
                self._make_cube(0, 0, 0),
                self._make_cube(0, 1, 1),
                self._make_cube(0, 2, 2),
                self._make_cube(1, 0, 3),
                self._make_cube(1, 1, 4),
                self._make_cube(1, 2, 5),
            ]
        )
        cube = cubes.merge()[0]
        self.assertCML(
            cube, ("merge", "time_triple_merging4.cml"), checksum=False
        )

        cube = iris.cube.CubeList(cubes[:-1]).merge()[0]
        self.assertCML(
            cube, ("merge", "time_triple_merging5.cml"), checksum=False
        )


class TestCubeMergeTheoretical(tests.IrisTest):
    def test_simple_bounds_merge(self):
        cube1 = iris.tests.stock.simple_2d()
        cube2 = iris.tests.stock.simple_2d()

        cube1.add_aux_coord(
            DimCoord(np.int32(10), long_name="pressure", units="Pa")
        )
        cube2.add_aux_coord(
            DimCoord(np.int32(11), long_name="pressure", units="Pa")
        )

        r = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertCML(r, ("cube_merge", "test_simple_bound_merge.cml"))

    def test_simple_multidim_merge(self):
        cube1 = iris.tests.stock.simple_2d_w_multidim_coords()
        cube2 = iris.tests.stock.simple_2d_w_multidim_coords()

        cube1.add_aux_coord(
            DimCoord(np.int32(10), long_name="pressure", units="Pa")
        )
        cube2.add_aux_coord(
            DimCoord(np.int32(11), long_name="pressure", units="Pa")
        )

        r = iris.cube.CubeList([cube1, cube2]).merge()[0]
        self.assertCML(r, ("cube_merge", "multidim_coord_merge.cml"))

        # try transposing the cubes first
        cube1.transpose([1, 0])
        cube2.transpose([1, 0])
        r = iris.cube.CubeList([cube1, cube2]).merge()[0]
        self.assertCML(r, ("cube_merge", "multidim_coord_merge_transpose.cml"))

    def test_simple_points_merge(self):
        cube1 = iris.tests.stock.simple_2d(with_bounds=False)
        cube2 = iris.tests.stock.simple_2d(with_bounds=False)

        cube1.add_aux_coord(
            DimCoord(np.int32(10), long_name="pressure", units="Pa")
        )
        cube2.add_aux_coord(
            DimCoord(np.int32(11), long_name="pressure", units="Pa")
        )

        r = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertCML(r, ("cube_merge", "test_simple_merge.cml"))

        # check that the unique merging raises a Duplicate data error
        self.assertRaises(
            iris.exceptions.DuplicateDataError,
            iris.cube.CubeList([cube1, cube1]).merge,
            unique=True,
        )

        # check that non unique merging returns both cubes
        r = iris.cube.CubeList([cube1, cube1]).merge(unique=False)
        self.assertCML(r[0], ("cube_merge", "test_orig_point_cube.cml"))
        self.assertCML(r[1], ("cube_merge", "test_orig_point_cube.cml"))

        # test attribute merging
        cube1.attributes["my_attr1"] = "foo"
        r = iris.cube.CubeList([cube1, cube2]).merge()
        # result should be 2 cubes
        self.assertCML(r, ("cube_merge", "test_simple_attributes1.cml"))

        cube2.attributes["my_attr1"] = "bar"
        r = iris.cube.CubeList([cube1, cube2]).merge()
        # result should be 2 cubes
        self.assertCML(r, ("cube_merge", "test_simple_attributes2.cml"))

        cube2.attributes["my_attr1"] = "foo"
        r = iris.cube.CubeList([cube1, cube2]).merge()
        # result should be 1 cube
        self.assertCML(r, ("cube_merge", "test_simple_attributes3.cml"))


class TestCubeMergeWithAncils(tests.IrisTest):
    def _makecube(self, y, cm=False, av=False):
        cube = iris.cube.Cube([0, 0])
        cube.add_dim_coord(iris.coords.DimCoord([0, 1], long_name="x"), 0)
        cube.add_aux_coord(iris.coords.DimCoord(y, long_name="y"))
        if cm:
            cube.add_cell_measure(
                iris.coords.CellMeasure([1, 1], long_name="foo"), 0
            )
        if av:
            cube.add_ancillary_variable(
                iris.coords.AncillaryVariable([1, 1], long_name="bar"), 0
            )
        return cube

    def test_fail_missing_cell_measure(self):
        cube1 = self._makecube(0, cm=True)
        cube2 = self._makecube(1)
        cubes = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 2)

    def test_fail_missing_ancillary_variable(self):
        cube1 = self._makecube(0, av=True)
        cube2 = self._makecube(1)
        cubes = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 2)

    def test_fail_different_cell_measure(self):
        cube1 = self._makecube(0, cm=True)
        cube2 = self._makecube(1)
        cube2.add_cell_measure(
            iris.coords.CellMeasure([2, 2], long_name="foo"), 0
        )
        cubes = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 2)

    def test_fail_different_ancillary_variable(self):
        cube1 = self._makecube(0, av=True)
        cube2 = self._makecube(1)
        cube2.add_ancillary_variable(
            iris.coords.AncillaryVariable([2, 2], long_name="bar"), 0
        )
        cubes = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 2)

    def test_merge_with_cell_measure(self):
        cube1 = self._makecube(0, cm=True)
        cube2 = self._makecube(1, cm=True)
        cubes = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 1)
        self.assertEqual(cube1.cell_measures(), cubes[0].cell_measures())

    def test_merge_with_ancillary_variable(self):
        cube1 = self._makecube(0, av=True)
        cube2 = self._makecube(1, av=True)
        cubes = iris.cube.CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 1)
        self.assertEqual(
            cube1.ancillary_variables(), cubes[0].ancillary_variables()
        )

    def test_cell_measure_error_msg(self):
        msg = "cube.cell_measures differ"
        cube1 = self._makecube(0, cm=True)
        cube2 = self._makecube(1)
        with self.assertRaisesRegex(iris.exceptions.MergeError, msg):
            _ = iris.cube.CubeList([cube1, cube2]).merge_cube()

    def test_ancillary_variable_error_msg(self):
        msg = "cube.ancillary_variables differ"
        cube1 = self._makecube(0, av=True)
        cube2 = self._makecube(1)
        with self.assertRaisesRegex(iris.exceptions.MergeError, msg):
            _ = iris.cube.CubeList([cube1, cube2]).merge_cube()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris.fileformats.name_loaders.__calc_integration_period`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import datetime

from iris.fileformats.name_loaders import _calc_integration_period


class Test(tests.IrisTest):
    def test_30_min_av(self):
        time_avgs = ["             30min average"]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (30 * 60))]
        self.assertEqual(result, expected)

    def test_30_min_av_rspace(self):
        time_avgs = ["             30min average  "]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (30 * 60))]
        self.assertEqual(result, expected)

    def test_30_min_av_lstrip(self):
        time_avgs = ["             30min average".lstrip()]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (30 * 60))]
        self.assertEqual(result, expected)

    def test_3_hour_av(self):
        time_avgs = ["          3hr 0min average"]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (3 * 60 * 60))]
        self.assertEqual(result, expected)

    def test_3_hour_int(self):
        time_avgs = ["         3hr 0min integral"]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (3 * 60 * 60))]
        self.assertEqual(result, expected)

    def test_12_hour_av(self):
        time_avgs = ["         12hr 0min average"]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (12 * 60 * 60))]
        self.assertEqual(result, expected)

    def test_5_day_av(self):
        time_avgs = ["    5day 0hr 0min integral"]
        result = _calc_integration_period(time_avgs)
        expected = [datetime.timedelta(0, (5 * 24 * 60 * 60))]
        self.assertEqual(result, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris.io` package."""



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test set up of limited area map extents which bridge the date line.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import iris

# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import matplotlib.pyplot as plt
    from iris.plot import pcolormesh


@tests.skip_plot
@tests.skip_data
class TestExtent(tests.IrisTest):
    def test_dateline(self):
        dpath = tests.get_data_path(["PP", "nzgust.pp"])
        cube = iris.load_cube(dpath)
        pcolormesh(cube)
        # Ensure that the limited area expected for NZ is set.
        # This is set in longitudes with the datum set to the
        # International Date Line.
        self.assertTrue(
            -10 < plt.gca().get_xlim()[0] < -5
            and 5 < plt.gca().get_xlim()[1] < 10
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the Fieldsfile file loading plugin and FFHeader.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import collections

import numpy as np

import iris
import iris.fileformats._ff as ff
import iris.fileformats.pp as pp


class TestFF_HEADER(tests.IrisTest):
    def test_initialisation(self):
        self.assertEqual(ff.FF_HEADER[0], ("data_set_format_version", (0,)))
        self.assertEqual(ff.FF_HEADER[17], ("integer_constants", (99, 100)))

    def test_size(self):
        self.assertEqual(len(ff.FF_HEADER), 31)


@tests.skip_data
class TestFFHeader(tests.IrisTest):
    def setUp(self):
        self.filename = tests.get_data_path(("FF", "n48_multi_field"))
        self.ff_header = ff.FFHeader(self.filename)
        self.valid_headers = (
            "integer_constants",
            "real_constants",
            "level_dependent_constants",
            "lookup_table",
            "data",
        )
        self.invalid_headers = (
            "row_dependent_constants",
            "column_dependent_constants",
            "fields_of_constants",
            "extra_constants",
            "temp_historyfile",
            "compressed_field_index1",
            "compressed_field_index2",
            "compressed_field_index3",
        )

    def test_constructor(self):
        # Test FieldsFile header attribute lookup.
        self.assertEqual(self.ff_header.data_set_format_version, 20)
        self.assertEqual(self.ff_header.sub_model, 1)
        self.assertEqual(self.ff_header.vert_coord_type, 5)
        self.assertEqual(self.ff_header.horiz_grid_type, 0)
        self.assertEqual(self.ff_header.dataset_type, 3)
        self.assertEqual(self.ff_header.run_identifier, 0)
        self.assertEqual(self.ff_header.experiment_number, -32768)
        self.assertEqual(self.ff_header.calendar, 1)
        self.assertEqual(self.ff_header.grid_staggering, 3)
        self.assertEqual(self.ff_header.time_type, -32768)
        self.assertEqual(self.ff_header.projection_number, -32768)
        self.assertEqual(self.ff_header.model_version, 802)
        self.assertEqual(self.ff_header.obs_file_type, -32768)
        self.assertEqual(self.ff_header.last_fieldop_type, -32768)
        self.assertEqual(
            self.ff_header.first_validity_time, (2011, 7, 10, 18, 0, 0, 191)
        )
        self.assertEqual(
            self.ff_header.last_validity_time, (2011, 7, 10, 21, 0, 0, 191)
        )
        self.assertEqual(
            self.ff_header.misc_validity_time,
            (2012, 4, 30, 18, 12, 13, -32768),
        )
        self.assertEqual(self.ff_header.integer_constants.shape, (46,))
        self.assertEqual(self.ff_header.real_constants.shape, (38,))
        self.assertEqual(
            self.ff_header.level_dependent_constants.shape, (71, 8)
        )
        self.assertIsNone(self.ff_header.row_dependent_constants)
        self.assertIsNone(self.ff_header.column_dependent_constants)
        self.assertIsNone(self.ff_header.fields_of_constants)
        self.assertIsNone(self.ff_header.extra_constants)
        self.assertIsNone(self.ff_header.temp_historyfile)
        self.assertIsNone(self.ff_header.compressed_field_index1)
        self.assertIsNone(self.ff_header.compressed_field_index2)
        self.assertIsNone(self.ff_header.compressed_field_index3)
        self.assertEqual(self.ff_header.lookup_table, (909, 64, 5))
        self.assertEqual(self.ff_header.total_prognostic_fields, 3119)
        self.assertEqual(self.ff_header.data, (2049, 2961, -32768))

    def test_str(self):
        self.assertString(str(self.ff_header), ("FF", "ffheader.txt"))

    def test_repr(self):
        target = "FFHeader('" + self.filename + "')"
        self.assertEqual(repr(self.ff_header), target)

    def test_shape(self):
        self.assertEqual(self.ff_header.shape("data"), (2961, -32768))


@tests.skip_data
class TestFF2PP2Cube(tests.IrisTest):
    def setUp(self):
        self.filename = tests.get_data_path(("FF", "n48_multi_field"))

    def test_unit_pass_0(self):
        # Test FieldsFile to PPFields cube load.
        cube_by_name = collections.defaultdict(int)
        cubes = iris.load(self.filename)
        while cubes:
            cube = cubes.pop(0)
            standard_name = cube.standard_name
            cube_by_name[standard_name] += 1
            filename = "{}_{}.cml".format(
                standard_name, cube_by_name[standard_name]
            )
            self.assertCML(cube, ("FF", filename))

    def test_raw_to_table_count(self):
        filename = tests.get_data_path(("FF", "n48_multi_field_table_count"))
        cubes = iris.load_raw(filename)
        ff_header = ff.FFHeader(filename)
        table_count = ff_header.lookup_table[2]
        self.assertEqual(len(cubes), table_count)


@tests.skip_data
class TestFFieee32(tests.IrisTest):
    def test_iris_loading(self):
        ff32_fname = tests.get_data_path(("FF", "n48_multi_field.ieee32"))
        ff64_fname = tests.get_data_path(("FF", "n48_multi_field"))

        ff32_cubes = iris.load(ff32_fname)
        ff64_cubes = iris.load(ff64_fname)

        for ff32, ff64 in zip(ff32_cubes, ff64_cubes):
            # load the data
            _, _ = ff32.data, ff64.data
            self.assertEqual(ff32, ff64)


@tests.skip_data
class TestFFVariableResolutionGrid(tests.IrisTest):
    def setUp(self):
        self.filename = tests.get_data_path(("FF", "n48_multi_field"))

        self.ff2pp = ff.FF2PP(self.filename)
        self.ff_header = self.ff2pp._ff_header

        data_shape = (73, 96)
        delta = np.sin(np.linspace(0, np.pi * 5, data_shape[1])) * 5
        lons = np.linspace(0, 180, data_shape[1]) + delta
        lons = np.vstack([lons[:-1], lons[:-1] + 0.5 * np.diff(lons)]).T
        lons = np.reshape(lons, lons.shape, order="F")

        delta = np.sin(np.linspace(0, np.pi * 5, data_shape[0])) * 5
        lats = np.linspace(-90, 90, data_shape[0]) + delta
        lats = np.vstack([lats[:-1], lats[:-1] + 0.5 * np.diff(lats)]).T
        lats = np.reshape(lats, lats.shape, order="F")

        self.ff_header.column_dependent_constants = lons
        self.ff_header.row_dependent_constants = lats

        self.U_grid_x = lons[:-1, 1]
        self.V_grid_y = lats[:-1, 1]
        self.P_grid_x = lons[:, 0]
        self.P_grid_y = lats[:, 0]

        self.orig_make_pp_field = pp.make_pp_field

        def new_make_pp_field(header):
            field = self.orig_make_pp_field(header)
            field.stash = self.ff2pp._custom_stash
            field.bdx = field.bdy = field.bmdi
            return field

        # Replace the pp module function with this new function;
        # this gets called in PP2FF.
        pp.make_pp_field = new_make_pp_field

    def tearDown(self):
        pp.make_pp_field = self.orig_make_pp_field

    def _check_stash(self, stash, x_coord, y_coord):
        self.ff2pp._custom_stash = stash
        field = next(iter(self.ff2pp))
        self.assertArrayEqual(
            x_coord,
            field.x,
            ("x_coord was incorrect for " "stash {}".format(stash)),
        )
        self.assertArrayEqual(
            y_coord,
            field.y,
            ("y_coord was incorrect for " "stash {}".format(stash)),
        )

    def test_p(self):
        self._check_stash("m01s00i001", self.P_grid_x, self.P_grid_y)

    def test_u(self):
        self._check_stash("m01s00i002", self.U_grid_x, self.P_grid_y)

    def test_v(self):
        self._check_stash("m01s00i003", self.P_grid_x, self.V_grid_y)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.experimental.raster.export_geotiff` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import re

try:
    from osgeo import gdal
    from iris.experimental.raster import export_geotiff
except ImportError:
    gdal = None

from iris.coord_systems import GeogCS
from iris.coords import DimCoord
from iris.cube import Cube


@tests.skip_gdal
class TestDtypeAndValues(tests.IrisTest):
    def _cube(self, dtype):
        data = np.arange(12).reshape(3, 4).astype(dtype) + 20
        cube = Cube(data, "air_pressure_anomaly")
        coord = DimCoord(np.arange(3), "latitude", units="degrees")
        coord.guess_bounds()
        cube.add_dim_coord(coord, 0)
        coord = DimCoord(np.arange(4), "longitude", units="degrees")
        coord.guess_bounds()
        cube.add_dim_coord(coord, 1)
        return cube

    def _check_dtype(self, dtype, gdal_dtype):
        cube = self._cube(dtype)
        with self.temp_filename(".tif") as temp_filename:
            export_geotiff(cube, temp_filename)
            dataset = gdal.Open(temp_filename, gdal.GA_ReadOnly)
            band = dataset.GetRasterBand(1)
            self.assertEqual(band.DataType, gdal_dtype)
            self.assertEqual(band.ComputeRasterMinMax(1), (20, 31))

    def test_int16(self):
        self._check_dtype("<i2", gdal.GDT_Int16)

    def test_int16_big_endian(self):
        self._check_dtype(">i2", gdal.GDT_Int16)

    def test_int32(self):
        self._check_dtype("<i4", gdal.GDT_Int32)

    def test_int32_big_endian(self):
        self._check_dtype(">i4", gdal.GDT_Int32)

    def test_uint8(self):
        self._check_dtype("u1", gdal.GDT_Byte)

    def test_uint16(self):
        self._check_dtype("<u2", gdal.GDT_UInt16)

    def test_uint16_big_endian(self):
        self._check_dtype(">u2", gdal.GDT_UInt16)

    def test_uint32(self):
        self._check_dtype("<u4", gdal.GDT_UInt32)

    def test_uint32_big_endian(self):
        self._check_dtype(">u4", gdal.GDT_UInt32)

    def test_float32(self):
        self._check_dtype("<f4", gdal.GDT_Float32)

    def test_float32_big_endian(self):
        self._check_dtype(">f4", gdal.GDT_Float32)

    def test_float64(self):
        self._check_dtype("<f8", gdal.GDT_Float64)

    def test_float64_big_endian(self):
        self._check_dtype(">f8", gdal.GDT_Float64)

    def test_invalid(self):
        cube = self._cube("i1")
        with self.assertRaises(ValueError):
            with self.temp_filename(".tif") as temp_filename:
                export_geotiff(cube, temp_filename)


@tests.skip_gdal
class TestProjection(tests.IrisTest):
    def _cube(self, ellipsoid=None):
        data = np.arange(12).reshape(3, 4).astype("u1")
        cube = Cube(data, "air_pressure_anomaly")
        coord = DimCoord(
            np.arange(3), "latitude", units="degrees", coord_system=ellipsoid
        )
        coord.guess_bounds()
        cube.add_dim_coord(coord, 0)
        coord = DimCoord(
            np.arange(4), "longitude", units="degrees", coord_system=ellipsoid
        )
        coord.guess_bounds()
        cube.add_dim_coord(coord, 1)
        return cube

    def test_no_ellipsoid(self):
        cube = self._cube()
        with self.temp_filename(".tif") as temp_filename:
            export_geotiff(cube, temp_filename)
            dataset = gdal.Open(temp_filename, gdal.GA_ReadOnly)
            self.assertEqual(dataset.GetProjection(), "")

    def test_sphere(self):
        cube = self._cube(GeogCS(6377000))
        with self.temp_filename(".tif") as temp_filename:
            export_geotiff(cube, temp_filename)
            dataset = gdal.Open(temp_filename, gdal.GA_ReadOnly)
            projection_string = dataset.GetProjection()
            # String has embedded floating point values,
            # Test with values to N decimal places, using a regular expression.
            re_pattern = (
                r'GEOGCS\["unknown",DATUM\["unknown",'
                r'SPHEROID\["unknown",637....,0\]\],PRIMEM\["Greenwich",0\],'
                r'UNIT\["degree",0.01745[0-9]*,AUTHORITY\["EPSG","9122"\]\],'
                r'AXIS\["Latitude",NORTH\],AXIS\["Longitude",EAST\]\]'
            )
            re_exp = re.compile(re_pattern)
            self.assertIsNotNone(
                re_exp.match(projection_string),
                "projection string {!r} does not match {!r}".format(
                    projection_string, re_pattern
                ),
            )

    def test_ellipsoid(self):
        cube = self._cube(GeogCS(6377000, 6360000))
        with self.temp_filename(".tif") as temp_filename:
            export_geotiff(cube, temp_filename)
            dataset = gdal.Open(temp_filename, gdal.GA_ReadOnly)
            projection_string = dataset.GetProjection()
            # String has embedded floating point values,
            # Test with values to N decimal places, using a regular expression.
            re_pattern = (
                r'GEOGCS\["unknown",DATUM\["unknown",'
                r'SPHEROID\["unknown",637....,375.117[0-9]*\]\],'
                r'PRIMEM\["Greenwich",0\],UNIT\["degree",0.01745[0-9]*,'
                r'AUTHORITY\["EPSG","9122"\]\],AXIS\["Latitude",NORTH\],'
                r'AXIS\["Longitude",EAST\]\]'
            )
            re_exp = re.compile(re_pattern)
            self.assertIsNotNone(
                re_exp.match(projection_string),
                "projection string {!r} does not match {!r}".format(
                    projection_string, re_pattern
                ),
            )


@tests.skip_gdal
class TestGeoTransform(tests.IrisTest):
    def test_(self):
        data = np.arange(12).reshape(3, 4).astype(np.uint8)
        cube = Cube(data, "air_pressure_anomaly")
        coord = DimCoord([30, 40, 50], "latitude", units="degrees")
        coord.guess_bounds()
        cube.add_dim_coord(coord, 0)
        coord = DimCoord([-10, -5, 0, 5], "longitude", units="degrees")
        coord.guess_bounds()
        cube.add_dim_coord(coord, 1)
        with self.temp_filename(".tif") as temp_filename:
            export_geotiff(cube, temp_filename)
            dataset = gdal.Open(temp_filename, gdal.GA_ReadOnly)
            self.assertEqual(
                dataset.GetGeoTransform(), (-12.5, 5, 0, 55, 0, -10)
            )


if __name__ == "__main__":
    tests.main()



import matplotlib.pyplot as plt

import iris
import iris.plot as iplt


fname = iris.sample_data_path("air_temp.pp")
temperature = iris.load_cube(fname)

# Take a 1d slice using array style indexing.
temperature_1d = temperature[5, :]

iplt.plot(temperature_1d)

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function
:func:`iris.analysis.cartography.rotate_grid_vectors`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest.mock import Mock, call as mock_call

import numpy as np

from iris.cube import Cube
from iris.tests.stock import sample_2d_latlons

from iris.analysis.cartography import rotate_grid_vectors


class TestRotateGridVectors(tests.IrisTest):
    def _check_angles_calculation(
        self, angles_in_degrees=True, nan_angles_mask=None
    ):
        # Check basic maths on a 2d latlon grid.
        u_cube = sample_2d_latlons(regional=True, transformed=True)
        u_cube.units = "ms-1"
        u_cube.rename("dx")
        u_cube.data[...] = 0
        v_cube = u_cube.copy()
        v_cube.name("dy")

        # Define 6 different vectors, repeated in each data row.
        in_vu = np.array([(0, 1), (2, -1), (-1, -1), (-3, 1), (2, 0), (0, 0)])
        in_angs = np.rad2deg(np.arctan2(in_vu[..., 0], in_vu[..., 1]))
        in_mags = np.sqrt(np.sum(in_vu * in_vu, axis=1))
        v_cube.data[...] = in_vu[..., 0]
        u_cube.data[...] = in_vu[..., 1]

        # Define 5 different test rotation angles, one for each data row.
        rotation_angles = np.array([0.0, -45.0, 135, -140.0, 90.0])
        ang_cube_data = np.broadcast_to(rotation_angles[:, None], u_cube.shape)
        ang_cube = u_cube.copy()
        if angles_in_degrees:
            ang_cube.units = "degrees"
        else:
            ang_cube.units = "radians"
            ang_cube_data = np.deg2rad(ang_cube_data)
        ang_cube.data[:] = ang_cube_data

        if nan_angles_mask is not None:
            ang_cube.data[nan_angles_mask] = np.nan

        # Rotate all vectors by all the given angles.
        result = rotate_grid_vectors(u_cube, v_cube, ang_cube)
        out_u, out_v = [cube.data for cube in result]

        # Check that vector magnitudes were unchanged.
        out_mags = np.sqrt(out_u * out_u + out_v * out_v)
        expect_mags = in_mags[None, :]
        self.assertArrayAllClose(out_mags, expect_mags)

        # Check that vector angles are all as expected.
        out_angs = np.rad2deg(np.arctan2(out_v, out_u))
        expect_angs = in_angs[None, :] + rotation_angles[:, None]
        ang_diffs = out_angs - expect_angs
        # Fix for null vectors, and +/-360 differences.
        ang_diffs[np.abs(out_mags) < 0.001] = 0.0
        ang_diffs = ang_diffs % 360.0
        # Check that any differences are very small.
        self.assertArrayAllClose(ang_diffs, 0.0)

        # Check that results are always masked arrays, masked at NaN angles.
        self.assertTrue(np.ma.isMaskedArray(out_u))
        self.assertTrue(np.ma.isMaskedArray(out_v))
        if nan_angles_mask is not None:
            self.assertArrayEqual(out_u.mask, nan_angles_mask)
            self.assertArrayEqual(out_v.mask, nan_angles_mask)

    def test_angles_calculation(self):
        self._check_angles_calculation()

    def test_angles_in_radians(self):
        self._check_angles_calculation(angles_in_degrees=False)

    def test_angles_from_grid(self):
        # Check it will gets angles from 'u_cube', and pass any kwargs on to
        # the angles routine.
        u_cube = sample_2d_latlons(regional=True, transformed=True)
        u_cube = u_cube[:2, :3]
        u_cube.units = "ms-1"
        u_cube.rename("dx")
        u_cube.data[...] = 1.0
        v_cube = u_cube.copy()
        v_cube.name("dy")
        v_cube.data[...] = 0.0

        # Setup a fake angles result from the inner call to 'gridcell_angles'.
        angles_result_data = np.array(
            [[0.0, 90.0, 180.0], [-180.0, -90.0, 270.0]]
        )
        angles_result_cube = Cube(angles_result_data, units="degrees")
        angles_kwargs = {"this": 2}
        angles_call_patch = self.patch(
            "iris.analysis._grid_angles.gridcell_angles",
            Mock(return_value=angles_result_cube),
        )

        # Call the routine.
        result = rotate_grid_vectors(
            u_cube, v_cube, grid_angles_kwargs=angles_kwargs
        )

        self.assertEqual(
            angles_call_patch.call_args_list, [mock_call(u_cube, this=2)]
        )

        out_u, out_v = [cube.data for cube in result]
        # Records what results should be for the various n*90deg rotations.
        expect_u = np.array([[1.0, 0.0, -1.0], [-1.0, 0.0, 0.0]])
        expect_v = np.array([[0.0, 1.0, 0.0], [0.0, -1.0, -1.0]])
        # Check results are as expected.
        self.assertArrayAllClose(out_u, expect_u)
        self.assertArrayAllClose(out_v, expect_v)

    def test_nan_vectors(self):
        bad_angle_points = np.zeros((5, 6), dtype=bool)
        bad_angle_points[2, 3] = True
        self._check_angles_calculation(nan_angles_mask=bad_angle_points)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :func:`iris.analysis.geometry.geometry_area_weights`
function.

 """

# Import iris.tests first so that some things can be initialised before
# importing anything else.

import iris.tests as tests
import iris.tests.stock as stock

import warnings
import numpy as np
import shapely.geometry

from iris.analysis.geometry import geometry_area_weights
from iris.coords import DimCoord
from iris.cube import Cube


class Test(tests.IrisTest):
    def setUp(self):
        x_coord = DimCoord([1.0, 3.0], "longitude", bounds=[[0, 2], [2, 4]])
        y_coord = DimCoord([1.0, 3.0], "latitude", bounds=[[0, 2], [2, 4]])
        self.data = np.empty((4, 2, 2))
        dim_coords_and_dims = [(y_coord, (1,)), (x_coord, (2,))]
        self.cube = Cube(self.data, dim_coords_and_dims=dim_coords_and_dims)
        self.geometry = shapely.geometry.Polygon(
            [(3, 3), (3, 50), (50, 50), (50, 3)]
        )

    def test_no_overlap(self):
        geometry = shapely.geometry.Polygon([(4, 4), (4, 6), (6, 6), (6, 4)])
        weights = geometry_area_weights(self.cube, geometry)
        self.assertEqual(np.sum(weights), 0)

    def test_overlap(self):
        weights = geometry_area_weights(self.cube, self.geometry)
        expected = np.repeat(
            [[[0.0, 0.0], [0.0, 1.0]]], self.data.shape[0], axis=0
        )
        self.assertArrayEqual(weights, expected)

    def test_overlap_normalize(self):
        weights = geometry_area_weights(
            self.cube, self.geometry, normalize=True
        )
        expected = np.repeat(
            [[[0.0, 0.0], [0.0, 0.25]]], self.data.shape[0], axis=0
        )
        self.assertArrayEqual(weights, expected)

    @tests.skip_data
    def test_distinct_xy(self):
        cube = stock.simple_pp()
        cube = cube[:4, :4]
        lon = cube.coord("longitude")
        lat = cube.coord("latitude")
        lon.guess_bounds()
        lat.guess_bounds()
        from iris.util import regular_step

        quarter = abs(regular_step(lon) * regular_step(lat) * 0.25)
        half = abs(regular_step(lon) * regular_step(lat) * 0.5)
        minx = 3.7499990463256836
        maxx = 7.499998092651367
        miny = 84.99998474121094
        maxy = 89.99998474121094
        geometry = shapely.geometry.box(minx, miny, maxx, maxy)
        weights = geometry_area_weights(cube, geometry)
        target = np.array(
            [
                [0, quarter, quarter, 0],
                [0, half, half, 0],
                [0, quarter, quarter, 0],
                [0, 0, 0, 0],
            ]
        )
        self.assertTrue(np.allclose(weights, target))

    @tests.skip_data
    def test_distinct_xy_bounds(self):
        # cases where geometry bnds are outside cube bnds correctly handled?
        cube = stock.simple_pp()
        cube = cube[:4, :4]
        lon = cube.coord("longitude")
        lat = cube.coord("latitude")
        lon.guess_bounds()
        lat.guess_bounds()
        from iris.util import regular_step

        quarter = abs(regular_step(lon) * regular_step(lat) * 0.25)
        half = abs(regular_step(lon) * regular_step(lat) * 0.5)
        full = abs(regular_step(lon) * regular_step(lat))
        minx = 3.7499990463256836
        maxx = 13.12499619
        maxx_overshoot = 15.0
        miny = 84.99998474121094
        maxy = 89.99998474121094
        geometry = shapely.geometry.box(minx, miny, maxx, maxy)
        geometry_overshoot = shapely.geometry.box(
            minx, miny, maxx_overshoot, maxy
        )
        weights = geometry_area_weights(cube, geometry)
        weights_overshoot = geometry_area_weights(cube, geometry_overshoot)
        target = np.array(
            [
                [0, quarter, half, half],
                [0, half, full, full],
                [0, quarter, half, half],
                [0, 0, 0, 0],
            ]
        )
        self.assertTrue(np.allclose(weights, target))
        self.assertTrue(np.allclose(weights_overshoot, target))

    @tests.skip_data
    def test_distinct_xy_bounds_pole(self):
        # is UserWarning issued for out-of-bounds? results will be unexpected!
        cube = stock.simple_pp()
        cube = cube[:4, :4]
        lon = cube.coord("longitude")
        lat = cube.coord("latitude")
        lon.guess_bounds()
        lat.guess_bounds()
        from iris.util import regular_step

        quarter = abs(regular_step(lon) * regular_step(lat) * 0.25)
        half = abs(regular_step(lon) * regular_step(lat) * 0.5)
        top_cell_half = abs(regular_step(lon) * (90 - lat.bounds[0, 1]) * 0.5)
        minx = 3.7499990463256836
        maxx = 7.499998092651367
        miny = 84.99998474121094
        maxy = 99.99998474121094
        geometry = shapely.geometry.box(minx, miny, maxx, maxy)
        # see http://stackoverflow.com/a/3892301 to assert warnings
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # always trigger all warnings
            weights = geometry_area_weights(cube, geometry)
            self.assertEqual(
                str(w[-1].message),
                "The geometry exceeds the "
                "cube's y dimension at the upper end.",
            )
            self.assertTrue(issubclass(w[-1].category, UserWarning))
        target = np.array(
            [
                [0, top_cell_half, top_cell_half, 0],
                [0, half, half, 0],
                [0, quarter, quarter, 0],
                [0, 0, 0, 0],
            ]
        )
        self.assertTrue(np.allclose(weights, target))

    def test_shared_xy(self):
        cube = stock.track_1d()
        geometry = shapely.geometry.box(1, 4, 3.5, 7)
        weights = geometry_area_weights(cube, geometry)
        target = np.array([0, 0, 2, 0.5, 0, 0, 0, 0, 0, 0, 0])
        self.assertTrue(np.allclose(weights, target))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:class:`iris.analysis.trajectory.UnstructuredNearestNeigbourRegridder`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.analysis.trajectory import (
    UnstructuredNearestNeigbourRegridder as unn_gridder,
)
from iris.coords import AuxCoord, DimCoord
from iris.coord_systems import GeogCS, RotatedGeogCS
from iris.cube import Cube, CubeList


class MixinExampleSetup:
    # Common code for regridder test classes.

    def setUp(self):
        # Basic test values.
        src_x_y_value = np.array(
            [
                [20.12, 11.73, 0.01],
                [120.23, -20.73, 1.12],
                [290.34, 33.88, 2.23],
                [-310.45, 57.8, 3.34],
            ]
        )
        tgt_grid_x = np.array([-173.2, -100.3, -32.5, 1.4, 46.6, 150.7])
        tgt_grid_y = np.array([-80.1, -30.2, 0.3, 47.4, 75.5])

        # Make sample 1-D source cube.
        src = Cube(src_x_y_value[:, 2])
        src.add_aux_coord(
            AuxCoord(
                src_x_y_value[:, 0], standard_name="longitude", units="degrees"
            ),
            0,
        )
        src.add_aux_coord(
            AuxCoord(
                src_x_y_value[:, 1], standard_name="latitude", units="degrees"
            ),
            0,
        )
        self.src_cube = src

        # Make sample grid cube.
        grid = Cube(np.zeros(tgt_grid_y.shape + tgt_grid_x.shape))
        grid.add_dim_coord(
            DimCoord(tgt_grid_y, standard_name="latitude", units="degrees"), 0
        )
        grid.add_dim_coord(
            DimCoord(tgt_grid_x, standard_name="longitude", units="degrees"), 1
        )
        self.grid_cube = grid

        # Make expected-result, from the expected source-index at each point.
        expected_result_indices = np.array(
            [
                [1, 1, 1, 1, 1, 1],
                [1, 2, 0, 0, 0, 1],
                [1, 2, 2, 0, 0, 1],
                [3, 2, 2, 3, 3, 3],
                [3, 2, 3, 3, 3, 3],
            ]
        )
        self.expected_data = self.src_cube.data[expected_result_indices]

        # Make a 3D source cube, based on the existing 2d test data.
        z_cubes = [src.copy() for _ in range(3)]
        for i_z, z_cube in enumerate(z_cubes):
            z_cube.add_aux_coord(DimCoord([i_z], long_name="z"))
            z_cube.data = z_cube.data + 100.0 * i_z
        self.src_z_cube = CubeList(z_cubes).merge_cube()

        # Make a corresponding 3d expected result.
        self.expected_data_zxy = self.src_z_cube.data[
            :, expected_result_indices
        ]

    def _check_expected(
        self,
        src_cube=None,
        grid_cube=None,
        expected_data=None,
        expected_coord_names=None,
    ):
        # Test regridder creation + operation against expected results.
        if src_cube is None:
            src_cube = self.src_cube
        if grid_cube is None:
            grid_cube = self.grid_cube
        gridder = unn_gridder(src_cube, grid_cube)
        result = gridder(src_cube)
        if expected_coord_names is not None:
            # Check result coordinate identities.
            self.assertEqual(
                [coord.name() for coord in result.coords()],
                expected_coord_names,
            )
        if expected_data is None:
            # By default, check against the 'standard' data result.
            expected_data = self.expected_data
        self.assertArrayEqual(result.data, expected_data)
        return result


class Test__init__(MixinExampleSetup, tests.IrisTest):
    # Exercise all the constructor argument checks.

    def test_fail_no_src_x(self):
        self.src_cube.remove_coord("longitude")
        msg_re = "Source cube must have X- and Y-axis coordinates"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_no_src_y(self):
        self.src_cube.remove_coord("latitude")
        msg_re = "Source cube must have X- and Y-axis coordinates"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_bad_src_dims(self):
        self.src_cube = self.grid_cube
        msg_re = "Source.*same cube dimensions"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_mixed_latlons(self):
        self.src_cube.coord("longitude").rename("projection_x_coordinate")
        msg_re = "any.*latitudes/longitudes.*all must be"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_bad_latlon_units(self):
        self.grid_cube.coord("longitude").units = "m"
        msg_re = 'does not convert to "degrees"'
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_non_latlon_units_mismatch(self):
        # Convert all to non-latlon system (does work: see in "Test__call__").
        for cube in (self.src_cube, self.grid_cube):
            for axis_name in ("x", "y"):
                coord = cube.coord(axis=axis_name)
                coord_name = "projection_{}_coordinate".format(axis_name)
                coord.rename(coord_name)
                coord.units = "m"
        # Change one of the output units.
        self.grid_cube.coord(axis="x").units = "1"
        msg_re = "Source and target.*must have the same units"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_no_tgt_x(self):
        self.grid_cube.remove_coord("longitude")
        msg_re = "must contain a single 1D x coordinate"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_no_tgt_y(self):
        self.grid_cube.remove_coord("latitude")
        msg_re = "must contain a single 1D y coordinate"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_src_cs_mismatch(self):
        cs = GeogCS(1000.0)
        self.src_cube.coord("latitude").coord_system = cs
        msg_re = "must all have the same coordinate system"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_tgt_cs_mismatch(self):
        cs = GeogCS(1000.0)
        self.grid_cube.coord("latitude").coord_system = cs
        msg_re = "x.*and y.*must have the same coordinate system"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)

    def test_fail_src_tgt_cs_mismatch(self):
        cs = GeogCS(1000.0)
        self.src_cube.coord("latitude").coord_system = cs
        self.src_cube.coord("longitude").coord_system = cs
        msg_re = "Source and target.*same coordinate system"
        with self.assertRaisesRegex(ValueError, msg_re):
            unn_gridder(self.src_cube, self.grid_cube)


class Test__call__(MixinExampleSetup, tests.IrisTest):
    # Test regridder operation and results.

    def test_basic_latlon(self):
        # Check a test operation.
        self._check_expected(
            expected_coord_names=["latitude", "longitude"],
            expected_data=self.expected_data,
        )

    def test_non_latlon(self):
        # Check different answer in cartesian coordinates (no wrapping, etc).
        # Convert to non-latlon system, with the same coord values.
        for cube in (self.src_cube, self.grid_cube):
            for axis_name in ("x", "y"):
                coord = cube.coord(axis=axis_name)
                coord_name = "projection_{}_coordinate".format(axis_name)
                coord.rename(coord_name)
                coord.units = "m"
        # Check for a somewhat different result.
        non_latlon_indices = np.array(
            [
                [3, 0, 0, 0, 1, 1],
                [3, 0, 0, 0, 0, 1],
                [3, 0, 0, 0, 0, 1],
                [3, 0, 0, 0, 0, 1],
                [3, 0, 0, 0, 0, 1],
            ]
        )
        expected_data = self.src_cube.data[non_latlon_indices]
        self._check_expected(expected_data=expected_data)

    def test_multidimensional_xy(self):
        # Recast the 4-point source cube as 2*2 : should yield the same result.
        co_x = self.src_cube.coord(axis="x")
        co_y = self.src_cube.coord(axis="y")
        new_src = Cube(self.src_cube.data.reshape((2, 2)))
        new_x_co = AuxCoord(
            co_x.points.reshape((2, 2)),
            standard_name="longitude",
            units="degrees",
        )
        new_y_co = AuxCoord(
            co_y.points.reshape((2, 2)),
            standard_name="latitude",
            units="degrees",
        )
        new_src.add_aux_coord(new_x_co, (0, 1))
        new_src.add_aux_coord(new_y_co, (0, 1))
        self._check_expected(src_cube=new_src)

    def test_transposed_grid(self):
        # Show that changing the order of the grid X and Y has no effect.
        new_grid_cube = self.grid_cube.copy()
        new_grid_cube.transpose((1, 0))
        # Check that the new grid is in (X, Y) order.
        self.assertEqual(
            [coord.name() for coord in new_grid_cube.coords()],
            ["longitude", "latitude"],
        )
        # Check that the result is the same, dimension order is still Y,X.
        self._check_expected(
            grid_cube=new_grid_cube,
            expected_coord_names=["latitude", "longitude"],
        )

    def test_compatible_source(self):
        # Check operation on data with different dimensions to the original
        # source cube for the regridder creation.
        gridder = unn_gridder(self.src_cube, self.grid_cube)
        result = gridder(self.src_z_cube)
        self.assertEqual(
            [coord.name() for coord in result.coords()],
            ["z", "latitude", "longitude"],
        )
        self.assertArrayEqual(result.data, self.expected_data_zxy)

    def test_fail_incompatible_source(self):
        # Check that a slightly modified source cube is *not* acceptable.
        modified_src_cube = self.src_cube.copy()
        points = modified_src_cube.coord(axis="x").points
        points[0] += 0.01
        modified_src_cube.coord(axis="x").points = points
        gridder = unn_gridder(self.src_cube, self.grid_cube)
        msg = "not defined on the same source grid"
        with self.assertRaisesRegex(ValueError, msg):
            gridder(modified_src_cube)

    def test_transposed_source(self):
        # Check operation on data where the 'trajectory' dimension is not the
        # last one.
        src_z_cube = self.src_z_cube
        src_z_cube.transpose((1, 0))
        self._check_expected(
            src_cube=src_z_cube, expected_data=self.expected_data_zxy
        )

    def test_radians_degrees(self):
        # Check source + target unit conversions, grid and result in degrees.
        for axis_name in ("x", "y"):
            self.src_cube.coord(axis=axis_name).convert_units("radians")
            self.grid_cube.coord(axis=axis_name).convert_units("degrees")
        result = self._check_expected()
        self.assertEqual(result.coord(axis="x").units, "degrees")

    def test_degrees_radians(self):
        # Check source + target unit conversions, grid and result in radians.
        for axis_name in ("x", "y"):
            self.src_cube.coord(axis=axis_name).convert_units("degrees")
            self.grid_cube.coord(axis=axis_name).convert_units("radians")
        result = self._check_expected()
        self.assertEqual(result.coord(axis="x").units, "radians")

    def test_alternative_cs(self):
        # Check the result is just the same in a different coordinate system.
        cs = RotatedGeogCS(
            grid_north_pole_latitude=75.3,
            grid_north_pole_longitude=102.5,
            ellipsoid=GeogCS(100.0),
        )
        for cube in (self.src_cube, self.grid_cube):
            for coord_name in ("longitude", "latitude"):
                cube.coord(coord_name).coord_system = cs
        self._check_expected()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._epoch_date_hours`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import cf_units
from cf_units import Unit
from cftime import datetime as nc_datetime

from iris.fileformats.pp_load_rules import (
    _epoch_date_hours as epoch_hours_call,
)


#
# Run tests for each of the possible calendars from PPfield.calendar().
# Test year=0 and all=0 cases, plus "normal" dates, for each calendar.
# Result values are the same as from 'date2num' in cftime version <= 1.0.1.
#


class TestEpochHours__gregorian(tests.IrisTest):
    def setUp(self):
        self.hrs_unit = Unit(
            "hours since epoch", calendar=cf_units.CALENDAR_GREGORIAN
        )

    def test_1970_1_1(self):
        test_date = nc_datetime(1970, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, 0.0)

    def test_ymd_1_1_1(self):
        test_date = nc_datetime(1, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17259936.0)

    def test_year_0(self):
        test_date = nc_datetime(0, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17268720.0)

    def test_ymd_0_0_0(self):
        test_date = nc_datetime(0, 0, 0)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17269488.0)

    def test_ymd_0_preserves_timeofday(self):
        hrs, mins, secs, usecs = (7, 13, 24, 335772)
        hours_in_day = (
            hrs + 1.0 / 60 * mins + 1.0 / 3600 * secs + (1.0e-6) / 3600 * usecs
        )
        test_date = nc_datetime(
            0, 0, 0, hour=hrs, minute=mins, second=secs, microsecond=usecs
        )
        result = epoch_hours_call(self.hrs_unit, test_date)
        # NOTE: the calculation is only accurate to approx +/- 0.5 seconds
        # in such a large number of hours -- even 0.1 seconds is too fine.
        absolute_tolerance = 0.5 / 3600
        self.assertArrayAllClose(
            result, -17269488.0 + hours_in_day, rtol=0, atol=absolute_tolerance
        )


class TestEpochHours__360day(tests.IrisTest):
    def setUp(self):
        self.hrs_unit = Unit(
            "hours since epoch", calendar=cf_units.CALENDAR_360_DAY
        )

    def test_1970_1_1(self):
        test_date = nc_datetime(1970, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, 0.0)

    def test_ymd_1_1_1(self):
        test_date = nc_datetime(1, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17012160.0)

    def test_year_0(self):
        test_date = nc_datetime(0, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17020800.0)

    def test_ymd_0_0_0(self):
        test_date = nc_datetime(0, 0, 0)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17021544.0)


class TestEpochHours__365day(tests.IrisTest):
    def setUp(self):
        self.hrs_unit = Unit(
            "hours since epoch", calendar=cf_units.CALENDAR_365_DAY
        )

    def test_1970_1_1(self):
        test_date = nc_datetime(1970, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, 0.0)

    def test_ymd_1_1_1(self):
        test_date = nc_datetime(1, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17248440.0)

    def test_year_0(self):
        test_date = nc_datetime(0, 1, 1)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17257200.0)

    def test_ymd_0_0_0(self):
        test_date = nc_datetime(0, 0, 0)
        result = epoch_hours_call(self.hrs_unit, test_date)
        self.assertEqual(result, -17257968.0)


class TestEpochHours__invalid_calendar(tests.IrisTest):
    def test_bad_calendar(self):
        # Setup a unit with an unrecognised calendar
        hrs_unit = Unit(
            "hours since epoch", calendar=cf_units.CALENDAR_ALL_LEAP
        )
        # Test against a date with year=0, which requires calendar correction.
        test_date = nc_datetime(0, 1, 1)
        # Check that this causes an error.
        with self.assertRaisesRegex(ValueError, "unrecognised calendar"):
            epoch_hours_call(hrs_unit, test_date)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Load, configure and register color map palettes and initialise
color map meta-data mappings.

"""

from functools import wraps
import os
import os.path
import re

import cf_units
import matplotlib.cm as mpl_cm
import matplotlib.colors as mpl_colors
import numpy as np

import iris.cube
import iris.config


# Symmetric normalization function pivot points by SI unit.
PIVOT_BY_UNIT = {cf_units.Unit("K"): 273.15}

# Color map names by palette file metadata field value.
CMAP_BREWER = set()
_CMAP_BY_SCHEME = None
_CMAP_BY_KEYWORD = None
_CMAP_BY_STD_NAME = None

_MISSING_KWARG_CMAP = "missing kwarg cmap"
_MISSING_KWARG_NORM = "missing kwarg norm"


def is_brewer(cmap):
    """
    Determine whether the color map is a Cynthia Brewer color map.

    Args:

    * cmap:
        The color map instance.

    Returns:
        Boolean.

    """
    result = False
    if cmap is not None:
        result = cmap.name in CMAP_BREWER
    return result


def _default_cmap_norm(args, kwargs):
    """
    This function injects default cmap and norm behavour into the keyword
    arguments, based on the cube referenced within the positional arguments.
    """
    cube = None

    # Find the single cube reference within the positional arguments.
    for arg in args:
        if isinstance(arg, iris.cube.Cube):
            cube = arg
            break

    # Find the keyword arguments of interest.
    colors = kwargs.get("colors", None)
    # cmap = None to disable default behaviour.
    cmap = kwargs.get("cmap", _MISSING_KWARG_CMAP)
    # norm = None to disable default behaviour.
    norm = kwargs.get("norm", _MISSING_KWARG_NORM)

    # Note that "colors" and "cmap" keywords are mutually exclusive.
    if colors is None and cube is not None:
        std_name = cube.standard_name.lower() if cube.standard_name else ""

        # Perform default "cmap" keyword behaviour.
        if cmap == _MISSING_KWARG_CMAP:
            # Check for an exact match against standard name.
            cmaps = _CMAP_BY_STD_NAME.get(std_name, set())

            if len(cmaps) == 0:
                # Check for a fuzzy match against a keyword.
                for keyword in _CMAP_BY_KEYWORD.keys():
                    if keyword in std_name:
                        cmaps.update(_CMAP_BY_KEYWORD[keyword])

            # Add default color map to keyword arguments.
            if len(cmaps):
                cmap = sorted(cmaps, reverse=True)[0]
                kwargs["cmap"] = mpl_cm.get_cmap(cmap)

        # Perform default "norm" keyword behaviour.
        if norm == _MISSING_KWARG_NORM:
            if "anomaly" in std_name:
                # Determine the pivot point.
                pivot = PIVOT_BY_UNIT.get(cube.units, 0)
                norm = SymmetricNormalize(pivot)
                kwargs["norm"] = norm

    return args, kwargs


def cmap_norm(cube):
    """
    Determine the default :class:`matplotlib.colors.LinearSegmentedColormap`
    and :class:`iris.palette.SymmetricNormalize` instances associated with
    the cube.

    Args:

    * cube (:class:`iris.cube.Cube`):
        Source cube to generate default palette from.

    Returns:
        Tuple of :class:`matplotlib.colors.LinearSegmentedColormap` and
        :class:`iris.palette.SymmetricNormalize`

    """
    args, kwargs = _default_cmap_norm((cube,), {})
    return kwargs.get("cmap"), kwargs.get("norm")


def auto_palette(func):
    """
    Decorator wrapper function to control the default behaviour of the
    matplotlib cmap and norm keyword arguments.

    Args:

    * func (callable):
        Callable function to be wrapped by the decorator.

    Returns:
        Closure wrapper function.

    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        """
        Closure wrapper function to provide default keyword argument
        behaviour.

        """
        # Update the keyword arguments with defaults.
        args, kwargs = _default_cmap_norm(args, kwargs)
        # Call the wrapped function and return its result.
        return func(*args, **kwargs)

    # Return the closure wrapper function.
    return wrapper_func


class SymmetricNormalize(mpl_colors.Normalize):
    """
    Provides a symmetric normalization class around a given pivot point.
    """

    def __init__(self, pivot, *args, **kwargs):
        self.pivot = pivot
        self._vmin = None
        self._vmax = None
        mpl_colors.Normalize.__init__(self, *args, **kwargs)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.pivot)

    def _update(self, val, update_min=True, update_max=True):
        # Update both _vmin and _vmax from given value.
        val_diff = np.abs(val - self.pivot)
        vmin_diff = np.abs(self._vmin - self.pivot) if self._vmin else 0.0
        vmax_diff = np.abs(self._vmax - self.pivot) if self._vmax else 0.0
        diff = max(val_diff, vmin_diff, vmax_diff)

        if update_min:
            self._vmin = self.pivot - diff
        if update_max:
            self._vmax = self.pivot + diff

    @property
    def vmin(self):
        return getattr(self, "_vmin")

    @vmin.setter
    def vmin(self, val):
        if val is None:
            self._vmin = None
        elif self._vmax is None:
            # Don't set _vmax, it'll stop matplotlib from giving us one.
            self._update(val, update_max=False)
        else:
            # Set both _vmin and _vmax from value
            self._update(val)

    @property
    def vmax(self):
        return getattr(self, "_vmax")

    @vmax.setter
    def vmax(self, val):
        if val is None:
            self._vmax = None
        elif self._vmin is None:
            # Don't set _vmin, it'll stop matplotlib from giving us one.
            self._update(val, update_min=False)
        else:
            # Set both _vmin and _vmax from value
            self._update(val)


def _load_palette():
    """
    Load, configure and register color map palettes and initialise
    color map metadata mappings.

    """
    # Reference these module level namespace variables.
    global CMAP_BREWER, _CMAP_BY_SCHEME, _CMAP_BY_KEYWORD, _CMAP_BY_STD_NAME

    _CMAP_BY_SCHEME = {}
    _CMAP_BY_KEYWORD = {}
    _CMAP_BY_STD_NAME = {}

    filenames = []

    # Identify all .txt color map palette files.
    for root, dirs, files in os.walk(iris.config.PALETTE_PATH):
        # Prune any .svn directory from the tree walk.
        if ".svn" in dirs:
            del dirs[dirs.index(".svn")]
        # Identify any target .txt color map palette files.
        filenames.extend(
            [
                os.path.join(root, filename)
                for filename in files
                if os.path.splitext(filename)[1] == ".txt"
            ]
        )

    for filename in filenames:
        # Default color map name based on the file base-name (case-SENSITIVE).
        cmap_name = os.path.splitext(os.path.basename(filename))[0]
        cmap_scheme = None
        cmap_keywords = []
        cmap_std_names = []
        cmap_type = None

        # Perform default color map interpolation for quantization
        # levels per primary color.
        interpolate_flag = True

        # Read the file header.
        with open(filename) as file_handle:
            header = filter(
                lambda line: re.match(r"^\s*#.*:\s+.*$", line),
                file_handle.readlines(),
            )

        # Extract the file header metadata.
        for line in header:
            line = line.replace("#", "", 1).split(":")
            head = line[0].strip().lower()
            body = line[1].strip()

            if head == "name":
                # Case-SENSITIVE.
                cmap_name = "brewer_{}".format(body)

            if head == "scheme":
                # Case-insensitive.
                cmap_scheme = body.lower()

            if head == "keyword":
                # Case-insensitive.
                keywords = [part.strip().lower() for part in body.split(",")]
                cmap_keywords.extend(keywords)

            if head == "std_name":
                # Case-insensitive.
                std_names = [part.strip().lower() for part in body.split(",")]
                cmap_std_names.extend(std_names)

            if head == "interpolate":
                # Case-insensitive.
                interpolate_flag = body.lower() != "off"

            if head == "type":
                # Case-insensitive.
                cmap_type = body.lower()

        # Integrity check for meta-data 'type' field.
        assert cmap_type is not None, (
            'Missing meta-data "type" keyword for color map file, "%s"'
            % filename
        )
        assert (
            cmap_type == "rgb"
        ), 'Invalid type [%s] for color map file "%s"' % (cmap_type, filename)

        # Update the color map look-up dictionaries.
        CMAP_BREWER.add(cmap_name)

        if cmap_scheme is not None:
            scheme_group = _CMAP_BY_SCHEME.setdefault(cmap_scheme, set())
            scheme_group.add(cmap_name)

        for keyword in cmap_keywords:
            keyword_group = _CMAP_BY_KEYWORD.setdefault(keyword, set())
            keyword_group.add(cmap_name)

        for std_name in cmap_std_names:
            std_name_group = _CMAP_BY_STD_NAME.setdefault(std_name, set())
            std_name_group.add(cmap_name)

        # Load palette data and create the associated color map.
        cmap_data = np.loadtxt(filename)
        # Ensure to restrict the number of RGB quantization levels to
        # prevent color map interpolation.

        if interpolate_flag:
            # Perform default color map interpolation for quantization
            # levels per primary color.
            cmap = mpl_colors.LinearSegmentedColormap.from_list(
                cmap_name, cmap_data
            )
        else:
            # Restrict quantization levels per primary color (turn-off
            # interpolation).
            # Typically used for Brewer color maps.
            cmap = mpl_colors.LinearSegmentedColormap.from_list(
                cmap_name, cmap_data, N=len(cmap_data)
            )

        # Register the color map for use.
        mpl_cm.register_cmap(cmap=cmap)


# Ensure to load the color map palettes.
_load_palette()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides the infrastructure to support lenient client/service behaviour.

"""

from collections.abc import Iterable
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps
from inspect import getmodule
import threading


__all__ = [
    "LENIENT",
    "Lenient",
]


#: Default _Lenient services global activation state.
_LENIENT_ENABLE_DEFAULT = True

#: Default Lenient maths feature state.
_LENIENT_MATHS_DEFAULT = True

#: Protected _Lenient internal non-client, non-service keys.
_LENIENT_PROTECTED = ("active", "enable")


def _lenient_client(*dargs, services=None):
    """
    Decorator that allows a client function/method to declare at runtime that
    it is executing and requires lenient behaviour from a prior registered
    lenient service function/method.

    This decorator supports being called with no arguments e.g.,

        @_lenient_client()
        def func():
            pass

    This is equivalent to using it as a simple naked decorator e.g.,

        @_lenient_client
        def func()
            pass

    Alternatively, this decorator supports the lenient client explicitly
    declaring the lenient services that it wishes to use e.g.,

        @_lenient_client(services=(service1, service2, ...)
        def func():
            pass

    Args:

    * dargs (tuple of callable):
        A tuple containing the callable lenient client function/method to be
        wrapped by the decorator. This is automatically populated by Python
        through the decorator interface. No argument requires to be manually
        provided.

    Kwargs:

    * services (callable or str or iterable of callable/str)
        Zero or more function/methods, or equivalent fully qualified string names, of
        lenient service function/methods.

    Returns:
        Closure wrapped function/method.

    """
    ndargs = len(dargs)

    if ndargs:
        assert (
            ndargs == 1
        ), f"Invalid lenient client arguments, expecting 1 got {ndargs}."
        assert callable(
            dargs[0]
        ), "Invalid lenient client argument, expecting a callable."

    assert not (
        ndargs and services
    ), "Invalid lenient client, got both arguments and keyword arguments."

    if ndargs:
        # The decorator has been used as a simple naked decorator.
        (func,) = dargs

        @wraps(func)
        def lenient_client_inner_naked(*args, **kwargs):
            """
            Closure wrapper function to register the wrapped function/method
            as active at runtime before executing it.

            """
            with _LENIENT.context(active=_qualname(func)):
                result = func(*args, **kwargs)
            return result

        result = lenient_client_inner_naked
    else:
        # The decorator has been called with None, zero or more explicit lenient services.
        if services is None:
            services = ()

        if isinstance(services, str) or not isinstance(services, Iterable):
            services = (services,)

        def lenient_client_outer(func):
            @wraps(func)
            def lenient_client_inner(*args, **kwargs):
                """
                Closure wrapper function to register the wrapped function/method
                as active at runtime before executing it.

                """
                with _LENIENT.context(*services, active=_qualname(func)):
                    result = func(*args, **kwargs)
                return result

            return lenient_client_inner

        result = lenient_client_outer

    return result


def _lenient_service(*dargs):
    """
    Decorator that allows a function/method to declare that it supports lenient
    behaviour as a service.

    Registration is at Python interpreter parse time.

    The decorator supports being called with no arguments e.g.,

        @_lenient_service()
        def func():
            pass

    This is equivalent to using it as a simple naked decorator e.g.,

        @_lenient_service
        def func():
            pass

    Args:

    * dargs (tuple of callable):
        A tuple containing the callable lenient service function/method to be
        wrapped by the decorator. This is automatically populated by Python
        through the decorator interface. No argument requires to be manually
        provided.

    Returns:
        Closure wrapped function/method.

    """
    ndargs = len(dargs)

    if ndargs:
        assert (
            ndargs == 1
        ), f"Invalid lenient service arguments, expecting 1 got {ndargs}."
        assert callable(
            dargs[0]
        ), "Invalid lenient service argument, expecting a callable."

    if ndargs:
        # The decorator has been used as a simple naked decorator.
        # Thus the (single) argument is a function to be wrapped.
        # We just register the argument function as a lenient service, and
        # return it unchanged
        (func,) = dargs

        _LENIENT.register_service(func)

        # This decorator registers 'func': the func itself is unchanged.
        result = func

    else:
        # The decorator has been called with no arguments.
        # Return a decorator, to apply to 'func' immediately following.
        def lenient_service_outer(func):
            _LENIENT.register_service(func)

            # Decorator registers 'func', but func itself is unchanged.
            return func

        result = lenient_service_outer

    return result


def _qualname(func):
    """
    Return the fully qualified function/method string name.

    Args:

    * func (callable):
        Callable function/method. Non-callable arguments are simply
        passed through.

    .. note::
        Inherited methods will be qualified with the base class that
        defines the method.

    """
    result = func
    if callable(func):
        module = getmodule(func)
        result = f"{module.__name__}.{func.__qualname__}"

    return result


class Lenient(threading.local):
    def __init__(self, **kwargs):
        """
        A container for managing the run-time lenient features and options.

        Kwargs:

        * kwargs (dict)
            Mapping of lenient key/value options to enable/disable. Note that,
            only the lenient "maths" options is available, which controls
            lenient/strict cube arithmetic.

        For example::

            Lenient(maths=False)

        Note that, the values of these options are thread-specific.

        """
        # Configure the initial default lenient state.
        self._init()

        if not kwargs:
            # If not specified, set the default behaviour of the maths lenient feature.
            kwargs = dict(maths=_LENIENT_MATHS_DEFAULT)

        # Configure the provided (or default) lenient features.
        for feature, state in kwargs.items():
            self[feature] = state

    def __contains__(self, key):
        return key in self.__dict__

    def __getitem__(self, key):
        if key not in self.__dict__:
            cls = self.__class__.__name__
            emsg = f"Invalid {cls!r} option, got {key!r}."
            raise KeyError(emsg)
        return self.__dict__[key]

    def __repr__(self):
        cls = self.__class__.__name__
        msg = f"{cls}(maths={self.__dict__['maths']!r})"
        return msg

    def __setitem__(self, key, value):
        cls = self.__class__.__name__

        if key not in self.__dict__:
            emsg = f"Invalid {cls!r} option, got {key!r}."
            raise KeyError(emsg)

        if not isinstance(value, bool):
            emsg = f"Invalid {cls!r} option {key!r} value, got {value!r}."
            raise ValueError(emsg)

        self.__dict__[key] = value
        # Toggle the (private) lenient behaviour.
        _LENIENT.enable = value

    def _init(self):
        """Configure the initial default lenient state."""
        # This is the only public supported lenient feature i.e., cube arithmetic
        self.__dict__["maths"] = None

    @contextmanager
    def context(self, **kwargs):
        """
        Return a context manager which allows temporary modification of the
        lenient option state within the scope of the context manager.

        On entry to the context manager, all provided keyword arguments are
        applied. On exit from the context manager, the previous lenient
        option state is restored.

        For example::
            with iris.common.Lenient.context(maths=False):
                pass

        """

        def configure_state(state):
            for feature, value in state.items():
                self[feature] = value

        # Save the original state.
        original_state = deepcopy(self.__dict__)

        # Configure the provided lenient features.
        configure_state(kwargs)

        try:
            yield
        finally:
            # Restore the original state.
            self.__dict__.clear()
            self._init()
            configure_state(original_state)


###############################################################################


class _Lenient(threading.local):
    def __init__(self, *args, **kwargs):
        """
        A container for managing the run-time lenient services and client
        options for pre-defined functions/methods.

        Args:

        * args (callable or str or iterable of callable/str)
            A function/method or fully qualified string name of the function/method
            acting as a lenient service.

        Kwargs:

        * kwargs (dict of callable/str or iterable of callable/str)
            Mapping of lenient client function/method, or fully qualified string name
            of the function/method, to one or more lenient service
            function/methods or fully qualified string name of function/methods.

        For example::

            _Lenient(service1, service2, client1=service1, client2=(service1, service2))

        Note that, the values of these options are thread-specific.

        """
        # The executing lenient client at runtime.
        self.__dict__["active"] = None
        # The global lenient services state activation switch.
        self.__dict__["enable"] = _LENIENT_ENABLE_DEFAULT

        for service in args:
            self.register_service(service)

        for client, services in kwargs.items():
            self.register_client(client, services)

    def __call__(self, func):
        """
        Determine whether it is valid for the function/method to provide a
        lenient service at runtime to the actively executing lenient client.

        Args:

        * func (callable or str):
            A function/method or fully qualified string name of the function/method.

        Returns:
            Boolean.

        """
        result = False
        if self.__dict__["enable"]:
            service = _qualname(func)
            if service in self and self.__dict__[service]:
                active = self.__dict__["active"]
                if active is not None and active in self:
                    services = self.__dict__[active]
                    if isinstance(services, str) or not isinstance(
                        services, Iterable
                    ):
                        services = (services,)
                    result = service in services
        return result

    def __contains__(self, name):
        name = _qualname(name)
        return name in self.__dict__

    def __getattr__(self, name):
        if name not in self.__dict__:
            cls = self.__class__.__name__
            emsg = f"Invalid {cls!r} option, got {name!r}."
            raise AttributeError(emsg)
        return self.__dict__[name]

    def __getitem__(self, name):
        name = _qualname(name)
        if name not in self.__dict__:
            cls = self.__class__.__name__
            emsg = f"Invalid {cls!r} option, got {name!r}."
            raise KeyError(emsg)
        return self.__dict__[name]

    def __repr__(self):
        cls = self.__class__.__name__
        width = len(cls) + 1
        kwargs = [
            "{}={!r}".format(name, self.__dict__[name])
            for name in sorted(self.__dict__.keys())
        ]
        joiner = ",\n{}".format(" " * width)
        return "{}({})".format(cls, joiner.join(kwargs))

    def __setitem__(self, name, value):
        name = _qualname(name)
        cls = self.__class__.__name__

        if name not in self.__dict__:
            emsg = f"Invalid {cls!r} option, got {name!r}."
            raise KeyError(emsg)

        if name == "active":
            value = _qualname(value)
            if not isinstance(value, str) and value is not None:
                emsg = f"Invalid {cls!r} option {name!r}, expected a registered {cls!r} client, got {value!r}."
                raise ValueError(emsg)
            self.__dict__[name] = value
        elif name == "enable":
            self.enable = value
        else:
            if isinstance(value, str) or callable(value):
                value = (value,)
            if isinstance(value, Iterable):
                value = tuple([_qualname(item) for item in value])
            self.__dict__[name] = value

    @contextmanager
    def context(self, *args, **kwargs):
        """
        Return a context manager which allows temporary modification of
        the lenient option state for the active thread.

        On entry to the context manager, all provided keyword arguments are
        applied. On exit from the context manager, the previous lenient option
        state is restored.

        For example::
            with iris._LENIENT.context(example_lenient_flag=False):
                # ... code that expects some non-lenient behaviour

        .. note::
            iris._LENIENT.example_lenient_flag does not exist and is
            provided only as an example.

        """

        def update_client(client, services):
            if client in self.__dict__:
                existing_services = self.__dict__[client]
            else:
                existing_services = ()

            self.__dict__[client] = tuple(set(existing_services + services))

        # Save the original state.
        original_state = deepcopy(self.__dict__)

        # Temporarily update the state with the kwargs first.
        for name, value in kwargs.items():
            self[name] = value

        # Get the active client.
        active = self.__dict__["active"]

        if args:
            # Update the client with the provided services.
            new_services = tuple([_qualname(arg) for arg in args])

            if active is None:
                # Ensure not to use "context" as the ephemeral name
                # of the context manager runtime "active" lenient client,
                # as this causes a namespace clash with this method
                # i.e., _Lenient.context, via _Lenient.__getattr__
                active = "__context"
                self.__dict__["active"] = active
                self.__dict__[active] = new_services
            else:
                # Append provided services to any pre-existing services of the active client.
                update_client(active, new_services)
        else:
            # Append previous ephemeral services (for non-specific client) to the active client.
            if (
                active is not None
                and active != "__context"
                and "__context" in self.__dict__
            ):
                new_services = self.__dict__["__context"]
                update_client(active, new_services)

        try:
            yield
        finally:
            # Restore the original state.
            self.__dict__.clear()
            self.__dict__.update(original_state)

    @property
    def enable(self):
        """Return the activation state of the lenient services."""
        return self.__dict__["enable"]

    @enable.setter
    def enable(self, state):
        """
        Set the activate state of the lenient services.

        Setting the state to `False` disables all lenient services, and
        setting the state to `True` enables all lenient services.

        Args:

        * state (bool):
            Activate state for lenient services.

        """
        if not isinstance(state, bool):
            cls = self.__class__.__name__
            emsg = f"Invalid {cls!r} option 'enable', expected a {type(True)!r}, got {state!r}."
            raise ValueError(emsg)
        self.__dict__["enable"] = state

    def register_client(self, func, services, append=False):
        """
        Add the provided mapping of lenient client function/method to
        required lenient service function/methods.

        Args:

        * func (callable or str):
            A client function/method or fully qualified string name of the
            client function/method.

        * services (callable or str or iterable of callable/str):
            One or more service function/methods or fully qualified string names
            of the required service function/method.

        Kwargs:

        * append (bool):
            If True, append the lenient services to any pre-registered lenient
            services for the provided lenient client. Default is False.

        """
        func = _qualname(func)
        cls = self.__class__.__name__

        if func in _LENIENT_PROTECTED:
            emsg = (
                f"Cannot register {cls!r} client. "
                f"Please rename your client to be something other than {func!r}."
            )
            raise ValueError(emsg)
        if isinstance(services, str) or not isinstance(services, Iterable):
            services = (services,)
        if not len(services):
            emsg = f"Require at least one {cls!r} client service."
            raise ValueError(emsg)
        services = tuple([_qualname(service) for service in services])
        if append:
            # The original provided service order is not significant. There is
            # no requirement to preserve it, so it's safe to sort.
            existing = self.__dict__[func] if func in self else ()
            services = tuple(sorted(set(existing) | set(services)))
        self.__dict__[func] = services

    def register_service(self, func):
        """
        Add the provided function/method as providing a lenient service and
        activate it.

        Args:

        * func (callable or str):
            A service function/method or fully qualified string name of the
            service function/method.

        """
        func = _qualname(func)
        if func in _LENIENT_PROTECTED:
            cls = self.__class__.__name__
            emsg = (
                f"Cannot register {cls!r} service. "
                f"Please rename your service to be something other than {func!r}."
            )
            raise ValueError(emsg)
        self.__dict__[func] = True

    def unregister_client(self, func):
        """
        Remove the provided function/method as a lenient client using lenient services.

        Args:

        * func (callable or str):
            A function/method of fully qualified string name of the function/method.

        """
        func = _qualname(func)
        cls = self.__class__.__name__

        if func in _LENIENT_PROTECTED:
            emsg = f"Cannot unregister {cls!r} client, as {func!r} is a protected {cls!r} option."
            raise ValueError(emsg)

        if func in self.__dict__:
            value = self.__dict__[func]
            if isinstance(value, bool):
                emsg = f"Cannot unregister {cls!r} client, as {func!r} is not a valid {cls!r} client."
                raise ValueError(emsg)
            del self.__dict__[func]
        else:
            emsg = f"Cannot unregister unknown {cls!r} client {func!r}."
            raise ValueError(emsg)

    def unregister_service(self, func):
        """
        Remove the provided function/method as providing a lenient service.

        Args:

        * func (callable or str):
            A function/method or fully qualified string name of the function/method.

        """
        func = _qualname(func)
        cls = self.__class__.__name__

        if func in _LENIENT_PROTECTED:
            emsg = f"Cannot unregister {cls!r} service, as {func!r} is a protected {cls!r} option."
            raise ValueError(emsg)

        if func in self.__dict__:
            value = self.__dict__[func]
            if not isinstance(value, bool):
                emsg = f"Cannot unregister {cls!r} service, as {func!r} is not a valid {cls!r} service."
                raise ValueError(emsg)
            del self.__dict__[func]
        else:
            emsg = f"Cannot unregister unknown {cls!r} service {func!r}."
            raise ValueError(emsg)


#: (Private) Instance that manages all Iris run-time lenient client and service options.
_LENIENT = _Lenient()

#: (Public) Instance that manages all Iris run-time lenient features.
LENIENT = Lenient()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris.tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from datetime import datetime
from fnmatch import fnmatch
from glob import glob
import os
import subprocess

import iris


LICENSE_TEMPLATE = """# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details."""

# Guess iris repo directory of Iris - realpath is used to mitigate against
# Python finding the iris package via a symlink.
IRIS_DIR = os.path.realpath(os.path.dirname(iris.__file__))
IRIS_INSTALL_DIR = os.path.dirname(os.path.dirname(IRIS_DIR))
DOCS_DIR = os.path.join(IRIS_INSTALL_DIR, "docs", "iris")
DOCS_DIR = iris.config.get_option("Resources", "doc_dir", default=DOCS_DIR)
exclusion = ["Makefile", "build"]
DOCS_DIRS = glob(os.path.join(DOCS_DIR, "*"))
DOCS_DIRS = [
    DOC_DIR
    for DOC_DIR in DOCS_DIRS
    if os.path.basename(DOC_DIR) not in exclusion
]
# Get a dirpath to the git repository : allow setting with an environment
# variable, so Travis can test for headers in the repo, not the installation.
IRIS_REPO_DIRPATH = os.environ.get("IRIS_REPO_DIR", IRIS_INSTALL_DIR)


class TestLicenseHeaders(tests.IrisTest):
    @staticmethod
    def whatchanged_parse(whatchanged_output):
        """
        Returns a generator of tuples of data parsed from
        "git whatchanged --pretty='TIME:%at". The tuples are of the form
        ``(filename, last_commit_datetime)``

        Sample input::

            ['TIME:1366884020', '',
             ':000000 100644 0000000... 5862ced... A\tlib/iris/cube.py']

        """
        dt = None
        for line in whatchanged_output:
            if not line.strip():
                continue
            elif line.startswith("TIME:"):
                dt = datetime.fromtimestamp(int(line[5:]))
            else:
                # Non blank, non date, line -> must be the lines
                # containing the file info.
                fname = " ".join(line.split("\t")[1:])
                yield fname, dt

    @staticmethod
    def last_change_by_fname():
        """
        Return a dictionary of all the files under git which maps to
        the datetime of their last modification in the git history.

        .. note::

            This function raises a ValueError if the repo root does
            not have a ".git" folder. If git is not installed on the system,
            or cannot be found by subprocess, an IOError may also be raised.

        """
        # Check the ".git" folder exists at the repo dir.
        if not os.path.isdir(os.path.join(IRIS_REPO_DIRPATH, ".git")):
            msg = "{} is not a git repository."
            raise ValueError(msg.format(IRIS_REPO_DIRPATH))

        # Call "git whatchanged" to get the details of all the files and when
        # they were last changed.
        output = subprocess.check_output(
            ["git", "whatchanged", "--pretty=TIME:%ct"], cwd=IRIS_REPO_DIRPATH
        )

        output = output.decode().split("\n")
        res = {}
        for fname, dt in TestLicenseHeaders.whatchanged_parse(output):
            if fname not in res or dt > res[fname]:
                res[fname] = dt

        return res

    def test_license_headers(self):
        exclude_patterns = (
            "setup.py",
            "noxfile.py",
            "build/*",
            "dist/*",
            "docs/iris/gallery_code/*/*.py",
            "docs/iris/src/developers_guide/documenting/*.py",
            "docs/iris/src/userguide/plotting_examples/*.py",
            "docs/iris/src/userguide/regridding_plots/*.py",
            "docs/iris/src/developers_guide/gitwash_dumper.py",
            "docs/iris/src/_build/*",
            "lib/iris/analysis/_scipy_interpolate.py",
            "lib/iris/fileformats/_pyke_rules/*",
        )

        try:
            last_change_by_fname = self.last_change_by_fname()
        except ValueError as err:
            # Caught the case where this is not a git repo.
            msg = (
                "Iris installation did not look like a git repo?"
                "\nERR = {}\n\n"
            )
            return self.skipTest(msg.format(str(err)))

        failed = False
        for fname, last_change in sorted(last_change_by_fname.items()):
            full_fname = os.path.join(IRIS_REPO_DIRPATH, fname)
            if (
                full_fname.endswith(".py")
                and os.path.isfile(full_fname)
                and not any(fnmatch(fname, pat) for pat in exclude_patterns)
            ):
                with open(full_fname) as fh:
                    content = fh.read()
                    if not content.startswith(LICENSE_TEMPLATE):
                        print(
                            "The file {} does not start with the required "
                            "license header.".format(fname)
                        )
                        failed = True

        if failed:
            raise ValueError("There were license header failures. See stdout.")


if __name__ == "__main__":
    tests.main()
